#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from collections import defaultdict
import networkx as nx

from itertools import product, tee, count
from heapq import heappush, heappop

from typing import Dict, Iterable, Mapping, TypeVar, AbstractSet, Any, List

from .. import extrema

from networkx.algorithms.approximation.steinertree import steiner_tree


def window(iterable, size):
    iters = tee(iterable, size)
    for i in range(1, size):
        for each in iters[i:]:
            next(each, None)
    return zip(*iters)


class SignalRouter:

    def route(self, G: nx.Graph,
              terminals: List,
              node_cost_fn,
              edge_cost_fn
              ) -> nx.Graph:
        pass


class AStarRouter(SignalRouter):

    def __init__(self, heuristic_fn=None):

        if heuristic_fn is not None:
            self.heuristic_fn = heuristic_fn
        else:
            self.heuristic_fn = lambda n: 0

    def route(self, G: nx.Graph,
              terminals: List,
              node_cost_fn,
              edge_cost_fn
              ) -> nx.Graph:
        return spanning_subtree(
            G,
            terminals,
            node_cost_fn,
            edge_cost_fn,
            heuristic_fn=self.heuristic_fn,
        )


class ApproxSteinerTreeRouter(SignalRouter):

    def route(self, G: nx.Graph,
              terminals: List,
              node_cost_fn,
              edge_cost_fn
              ) -> nx.Graph:

        ccs = [G.subgraph(c) for c in nx.connected_components(G)]

        subgraph = None
        for cc in ccs:
            if all(t in cc for t in terminals):
                subgraph = cc
                break
        assert subgraph is not None
        G2 = subgraph

        for a, b in G2.edges:
            w = 0
            if node_cost_fn is not None:
                w += node_cost_fn(a) + node_cost_fn(b)
            if edge_cost_fn is not None:
                w += edge_cost_fn((a, b))
            G2[a][b]['weight'] = w

        return steiner_tree(G2, terminals, weight='weight')


class DijkstraRouter(SignalRouter):

    def __init__(self):
        pass

    def route(self, G: nx.Graph,
              terminals: List[Any],
              node_cost_fn,
              edge_cost_fn
              ) -> nx.Graph:
        return spanning_subtree(
            G,
            terminals,
            node_cost_fn,
            edge_cost_fn,
        )


def spanning_subtree(
        G: nx.Graph,
        terminals: List,
        node_cost_fn,
        edge_cost_fn,
        heuristic_fn=None,
) -> nx.Graph:
    """ Creates a spanning subtree connecting all terminals.
    Assumes a connected graph G.

    Based on PathFinder and Prim's algorithm.

    Parameters
    ----------
    G: nx.Graph
            Routing graph
    terminal: List of terminal nodes
            Terminal nodes of signals to be connected.
    node_cost_fn: node -> number
    edge_cost_fn  node, node -> number
    heuristic_fn: function source, target -> number
            See `dijkstra_traverse` function
    """

    S = nx.Graph()
    #
    # center, center_dist = absolute_1_center(
    #     G,
    #     terminals,
    #     node_cost_fn,
    #     edge_cost_fn
    # )
    #
    # source = center
    source = terminals[0]

    sinks = set(terminals)
    sinks -= {source}

    S.add_node(source)

    while sinks:
        # Initialize priority queue with current routing tree at cost 0.

        sources = set(S.nodes)

        path = shortest_path(G,
                             sources,
                             sinks,
                             node_cost_fn,
                             edge_cost_fn,
                             heuristic_fn=heuristic_fn
                             )

        nx.add_path(S, path)
        end = path[-1]
        sinks.remove(end)

    assert nx.is_tree(S)
    return S


def absolute_1_center(
        G,
        terminals,
        node_cost_fn,
        edge_cost_fn
):
    # Create distance maps for all terminals.
    distances = {
        t: dijkstra(
            G,
            [t],
            node_cost_fn,
            edge_cost_fn
        )
        for t in terminals}

    distances_by_node = defaultdict(list)
    for t, dists in distances.items():
        for n, dist in dists.items():
            distances_by_node[n].append(dist)

    max_distances = {n: max(dists) for n, dists in distances_by_node.items()}
    # Use sum of squared distances as metric.
    # max_distances = {n: sum(map(lambda x: x ** 2, dists)) for n, dists in distances_by_node.items()}

    min_max_distance = min(max_distances.items(), key=lambda x: x[1])

    return min_max_distance


def dijkstra(
        G,
        sources,
        node_cost_fn,
        edge_cost_fn,
        terminals=None,
        closest_terminal_only=False,
        heuristic_fn=None
):
    """ Create a distance map from all nodes to the source.
    If `terminals` is not None, then the search stops as soon as all terminals are found.

    Parameters
    ----------
    sources: Nodes to start the search from.
    terminals: Destinations.
    slack_ration: Tradeoff between base cost and history/sharing cost. If set to 1, only the base cost is taken into account.
    max_paths: Maximum number of paths to find. If set to 1 only the shortest path from a source to a terminal node will be returned.
    closest_terminal_only: If set to True, the search will terminate as soon as the first terminal is found.
    heuristic_fn: See `dijkstra_traverse` function
    """

    sinks = None
    if terminals:
        sinks = set(terminals)

    def node_handler_fn(n):
        if closest_terminal_only:
            return False

        if sinks:
            sinks.remove(n)
            return len(sinks) > 0
        else:
            return True

    if terminals is None:
        def multi_target_heuristic(source):
            return 0
    else:
        def multi_target_heuristic(source):
            return min((heuristic_fn(source, t) for t in terminals))

    result = dijkstra_traverse(
        G,
        sources,
        node_cost_fn,
        edge_cost_fn,
        node_handler_fn,
        heuristic_fn=multi_target_heuristic
    )

    return result


def dijkstra_traverse(
        G,
        sources,
        node_cost_fn,
        edge_cost_fn,
        node_handler_fn,
        heuristic_fn=None
) -> Dict[Any, float]:
    """ Create a distance map from all nodes to the source.
    Nodes are visited in increasing distance order and passed to `node_handler_fn`. The search is continued
    as long as `node_handler_fn` returns `True` and there are unvisited nodes left.

    Parameters
    ----------
    :param sources: Nodes to start the search from.
    :param node_cost_fn: Node cost function. node -> cost
    :param edge_cost_fn: Edge cost function. (node, node) -> cost
    :param node_handler_fn: A function Node -> Bool
            Each node will be passed to this function in increasing distance order. The search will be aborted if the handler returns `False`.

    :param heuristic_fn: Source -> Estimated cost to reach target.
            Heuristic function to estimate the cost from a node to a target.
            Shortest paths are found as long as the heuristic does not overestimate costs.

    :return: Returns a dictionary like {node: distance to source, ...}.
    """

    if heuristic_fn is None:
        def h(n):
            return 0

        heuristic_fn = h

    class PQElement:
        def __init__(self, priority, value):
            self.priority = priority
            self.value = value

        def __cmp__(self, other):
            if self.priority < other.priority:
                return -1
            elif self.priority > other.priority:
                return 1
            return 0

        def __lt__(self, other):
            return self.priority < other.priority

        def __gt__(self, other):
            return self.priority > other.priority

        def as_tuple(self):
            return self.priority, self.value

    # Initialize priority queue with source node at cost 0.
    pq = [PQElement(node_cost_fn(n), n) for n in sources]

    c = count()
    result = dict()
    # result = {n: node_cost_fn(n) for n in sources}
    visited = set()
    enqueued = dict()
    n_nodes = len(G)
    # Storage for trace.
    prev_node = dict()
    while pq:
        # Continue search from lowest-cost node.
        cost_m, m = heappop(pq).as_tuple()

        if m in visited:
            continue
        visited.add(m)

        result[m] = cost_m

        if node_handler_fn is not None:
            if not node_handler_fn(m):
                break

        if len(visited) == n_nodes:
            break

        # Loop over fanout nodes
        for edge in G.edges(m, data=True):
            _, n, data = edge

            if n in visited:
                continue

            effective_cost = edge_cost_fn((m, n)) + node_cost_fn(n)

            # Get previous node if any.
            previous = prev_node.get(n, None)

            cost_n = cost_m + effective_cost

            # old_cost_h = enqueued.get(n, None)
            # if old_cost_h is not None:
            #     # h was already computed, no need to recompute it again.
            #     cost_old, h = old_cost_h
            #
            #     if cost_old <= cost_n:
            #         # We already have a better candidate.
            #         continue
            # else:
            #     h = heuristic_fn(n)

            h = heuristic_fn(n)
            heappush(pq, PQElement(cost_n + h, n))
            # Cache h
            # enqueued[n] = cost_n, h

            # Remember node if it is augmenting the path.
            if previous is not None:
                prev_node[n] = min(previous, (cost_n, m), key=lambda x: x[0])
            else:
                prev_node[n] = (cost_n, m)

    return result


def shortest_path(
        G,
        sources,
        terminals,
        node_cost_fn,
        edge_cost_fn,
        heuristic_fn=None
):
    """ Finds the shortest path from the source to one of the terminals.

    Based on PathFinder and Prim's algorithm.

    Parameters
    ----------
    sources: Nodes to start the search from.
    terminals: Destinations.
    """
    sinks = None
    if terminals:
        sinks = set(terminals)

    if heuristic_fn is None:
        multi_target_heuristic = None
    else:
        if terminals:
            def multi_target_heuristic(source):
                return min((heuristic_fn(source, t) for t in terminals))
        else:
            def multi_target_heuristic(source):
                return 0

    closest_terminal = []

    def node_handler_fn(n):
        if sinks:
            # Continue as long as n is not a sink.
            if n in sinks:
                closest_terminal.append(n)
                return False
            return True
        else:
            return True

    distance_map = dijkstra_traverse(
        G,
        sources,
        node_cost_fn,
        edge_cost_fn,
        node_handler_fn,
        heuristic_fn=multi_target_heuristic
    )

    # Find path from closest terminal to a sink.

    assert len(closest_terminal) == 1
    closest_terminal = closest_terminal[0]
    assert closest_terminal in sinks

    result = trace_back(G, distance_map, closest_terminal, sources)

    return result


N = TypeVar('N')


def trace_back(G: nx.Graph, distance_map: Mapping[N, int], source: N, targets: AbstractSet[N]):
    """ Find the shortest path from `source` to one of the targets
    by tracing back based on a distance map.

    Parameters
    ----------
    G: nx.Graph
    distance_map: Distance to the source.
    source: Start node.
    targets: End nodes.
    """

    targets = set(targets)

    current = source

    trace = [current]

    while current not in targets:
        neighbors = G.neighbors(current)
        neighbors = filter(lambda n: n in distance_map, neighbors)

        # neighbors = sorted(neighbors)
        closest_nodes = extrema.all_min(neighbors, key=lambda n: distance_map[n])
        # TODO: Which node is next if there are multiple closest nodes?
        next_node = closest_nodes[0]
        trace.append(next_node)
        current = next_node

    trace.reverse()

    return trace


def test_dijkstra_router():
    """
    Create a routing tree for a single on a mesh graph signal and plot it.
    :return:
    """
    import matplotlib.pyplot as plt

    # Construct the graph.
    G = nx.Graph()

    num_x = 10
    num_y = 10
    x = range(0, num_x)
    y = range(0, num_y)

    # Store drawing positions of vertices. For plotting only.
    pos = {}

    # Construct mesh
    for name, (x, y) in enumerate(product(x, y)):
        G.add_node((x, y))
        pos[(x, y)] = (x, y)

        w = 1

        if x < num_x - 1 and not (1 <= y < 5 and x == 4):
            G.add_edge((x, y), (x + 1, y), weight=w, orientation='h')

        if y < num_y - 1:
            G.add_edge((x, y), (x, y + 1), weight=w, orientation='v')

    G.add_edge((8, 0), (9, 0), multi_via=2)

    # Plot the mesh.
    nx.draw_networkx(G, pos=pos, node_color='gray', node_size=8, edge_color='lightgray', hold=True)

    # This are the terminals to be connected.
    terminals = [(0, 0), (8, 5), (7, 7), (6, 3), (8, 0), (1, 8), (3, 3), (8, 4)]

    # Find the routing tree.
    router = DijkstraRouter()
    tree = router.route(G, terminals, node_cost_fn=lambda x: 1, edge_cost_fn=lambda x: 1)
    assert nx.is_tree(tree), "Routing solution should be a tree!"
    for n in terminals:
        assert n in tree, "Terminal is not in routing tree!"

    # Plot the result.

    edges = list(tree.edges)
    nx.draw_networkx_edges(G, pos, edgelist=edges, width=4, edge_color='red')

    nx.draw_networkx_nodes(G, pos, nodelist=terminals, node_size=32, node_color='black')
    # edge_labels = {(a, b): "%.2f" % data.get('weight', 0) for (a, b, data) in G.edges(data=True)}
    # nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels)

    plt.draw()
    plt.show()



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import networkx as nx
from itertools import combinations, permutations, chain, product

from typing import List, Set

import logging

logger = logging.getLogger(__name__)


def construct_even_degree_graphs(G: nx.MultiGraph) -> List[nx.MultiGraph]:
    """ Construct all graphs of even degree by inserting a minimal number of virtual edges.
    :param G: A nx.MultiGraph
    :return: List[nx.MultiGraph]
        Returns a list of all graphs that can be constructed by inserting virtual edges.
    """

    assert isinstance(G, nx.MultiGraph), Exception("G must be a nx.MultiGraph.")

    if nx.is_empty(G):
        logger.debug("Graph is empty.")
    else:
        # nx.is_connected is not defined for empty graphs.
        if not nx.is_connected(G):
            logger.debug("Graph is not connected. Assuming there is a transmission gate.")

    # Find nodes with odd degree.
    odd_degree_nodes = [n for n, deg in G.degree if deg % 2 == 1]

    assert len(odd_degree_nodes) % 2 == 0

    if len(odd_degree_nodes) == 0:
        # All node degrees are already even. Nothing to do.
        assert nx.is_connected(G), Exception("G must be a connected graph.")
        return [G.copy()]

    """
    Finding all even degree graphs by inserting a minimal number of edges works as follows:
    * Find nodes of odd degree `odd_degree_nodes` and find all possible pairings of them.
        * Find all dual partitionings of `odd_degree_nodes`
        * For each partitioning find all pairings across the two partitions.
    """

    # Find all dual partitionings of `odd_degree_nodes`
    partitions2 = list(combinations(odd_degree_nodes, len(odd_degree_nodes) // 2))
    partitions_a = partitions2[:len(partitions2) // 2]
    partitions_b = partitions2[:len(partitions2) // 2 - 1:-1]

    # Assert that a and b are complementary
    for a, b in zip(partitions_a, partitions_b):
        assert set(chain(a, b)) == set(odd_degree_nodes)
        assert set(a) & set(b) == set(), "Partitions must be disjoint!"

    even_degree_graphs = []
    # For each partitioning ...
    for partition_a, partition_b in zip(partitions_a, partitions_b):
        # ... find all pairings across the two partitions.
        for partition_b_permutation in permutations(partition_b):
            G2 = G.copy()
            for a, b in zip(partition_a, partition_b_permutation):
                assert G2.degree(a) % 2 == 1
                assert G2.degree(b) % 2 == 1
                G2.add_edge(a, b)

            for n, deg in G2.degree:
                assert deg % 2 == 0

            if nx.is_connected(G2):
                # In presence of transmission gates the resulting graph might not be connected.
                # Only store the graphs that are connected.
                # TODO: This might not scale if there are many transmission gates => Handle transmission gates more efficiently.
                even_degree_graphs.append(G2)

    for g1, g2 in combinations(even_degree_graphs, 2):
        assert g1 != g2, "There should be no duplicates."

    return even_degree_graphs


def find_all_euler_tours(G: nx.MultiGraph, start_node=None, end_node=None, visited_edges: Set = None):
    """ Find some tour starting at `start_node`.
    If `end_node` is given the trace will end there. However, it will not be a full tour.

    Parameters
    ----------
    G: The graph

    start_node: Start of the tour.

    visited_edges: Edges that should not appear in the tour anymore.

    Returns
    -------
    Tour through G starting and ending at `start_node`.
    """
    for n, deg in G.degree():
        assert deg % 2 == 0, Exception("All nodes in G must have even degree.")

    tours = []

    if visited_edges is None:
        visited_edges = set()

    if start_node is None:
        # Deterministically choose some start node.
        start_node = min(G.nodes)

    if end_node is None:
        end_node = start_node

    edges = list(G.edges(start_node, keys=True))

    assert len(edges) > 0

    for e in edges:
        a, b, c = e
        # Normalize edge by sorting start and end.
        ao, bo = tuple(sorted((a, b)))
        e_norm = ao, bo, c
        if e_norm not in visited_edges:
            if len(visited_edges) == len(G.edges) - 1:
                # Last edge
                assert b == end_node
                tours.append([(a, b, c)])
            else:
                assert len(visited_edges) < len(G.edges) - 1
                visited_edges_sub = visited_edges | {e_norm}
                start_sub = b
                sub_tours = find_all_euler_tours(G, start_sub, end_node, visited_edges=visited_edges_sub)

                tours.extend([[(a, b, c)] + s for s in sub_tours])

    return tours


def multigraph_networkx2rust(G):
    """ Convert a networkx MultiGraph into a edge list that can be processed by pyo3-cell.
    """

    # Map nodes to indices
    nodemap = {n: i for i, n in enumerate(G.nodes)}

    rustgraph = nx.MultiGraph()
    rustgraph.add_nodes_from((nodemap[n] for n in G.nodes))
    rustgraph.add_edges_from(((nodemap[a], nodemap[b], d) for a, b, d in G.edges))

    rust_edges = list(rustgraph.edges)

    return rust_edges, nodemap


def multigraph_rust2networkx(rust_routing_trees, nodemap):
    """ Inverse transformation of `multigraph_networkx2rust`.
    """

    reverse_map = {v: k for k, v in nodemap.items()}

    # Convert back to python nodes.
    routing_edges = [
        [(reverse_map[a], reverse_map[b]) for a, b in rt]
        for rt in rust_routing_trees
    ]
    routing_trees = []
    for re in routing_edges:
        rt = nx.Graph()
        rt.add_edges_from(re)
        routing_trees.append(rt)

    return routing_trees


if __name__ == '__main__':
    import pyo3_cell

    G = nx.MultiGraph()
    G.add_edge('a', 'b')
    # G.add_edge('b','c')
    # G.add_edge('b','d')
    # G.add_edge('b','e')
    # G.add_edge('c','f')
    # G.add_edge('d','f')
    # G.add_edge('e','f')
    G.add_edge('f', 'a')
    G.add_edge('g', 'a')

    G = construct_even_degree_graphs(G)[0]

    # euler_tour = find_euler_tour(G)

    # print(euler_tour)

    all_tours = find_all_euler_tours(G, 'a', 'a')

    rust_graph, node_map = multigraph_networkx2rust(G)
    print(rust_graph)
    all_tours_pyo3 = pyo3_cell.find_all_euler_tours(rust_graph, node_map['a'])

    print(len(all_tours))
    print(len(all_tours_pyo3))
    print(all_tours)
    print()
    print(all_tours_pyo3)



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from itertools import product
import numpy as np
from typing import Sequence, Tuple, Iterator


class Grid:

    def __init__(self, start: Tuple[int, int], end: Tuple[int, int], step: Tuple[int, int]):
        self.start = start
        self.end = end
        self.step = step

        self._dimension = len(start)

        assert all([len(v) == self._dimension for v in (start, end, step)]), \
            "start, end and step must have the same dimensions."

    def __iter__(self) -> Iterator[Tuple[int, int]]:
        """ Get iterator over grid points.
        :return:
        """

        return product(*(range(a, b, c)
                         for a, b, c in zip(self.start, self.end, self.step)
                         ))

    def meshgrid(self):
        """ Return the grid points as a numpy meshgrid.
        :return:
        """
        return np.meshgrid(*(np.arange(a, b, c)
                             for a, b, c in zip(self.start, self.end, self.step)
                             ))

    def _is_on_grid_axis(self, point: Sequence[int], dimension: int) -> bool:
        """ Check if `point` is aligned with grid regarding one dimension.
        :param point:
        :param dimension:
        :return:
        """
        x = point[dimension]
        start = self.start[dimension]
        step = self.step[dimension]
        end = self.end[dimension]

        return (x - start) % step == 0 and start <= x < end

    def is_on_grid(self, point) -> bool:
        return all((self._is_on_grid_axis(point, d) for d in range(self._dimension)))

    def _grid_floor(self, x: int, dimension: int) -> int:
        """ Round a coordinate down to next grid coordinate.
        """
        start = self.start[dimension]
        step = self.step[dimension]
        x = int(x)
        x = (x - start) // step * step + start

        return x

    def _grid_ceil(self, x: int, dimension: int) -> int:
        """ Round a coordinate up to next grid coordinate.
        """
        return self._grid_floor(x + self.step[dimension] - 1, dimension)

    def _grid_round(self, x: int, dimension: int) -> int:
        return self._grid_floor(x + self.step[dimension] // 2, dimension)


class Grid2D(Grid):

    def __init__(self, start: Tuple[int, int], end: Tuple[int, int], step: Tuple[int, int]):
        super().__init__(start, end, step)
        assert self._dimension == 2, "Dimension must be 2."

    def grid_floor_x(self, point):
        x, y = point
        return self._grid_floor(x, 0), y

    def grid_floor_y(self, point):
        x, y = point
        return x, self._grid_floor(y, 1)

    def grid_floor_xy(self, point):
        x, y = point
        return self._grid_floor(x, 0), self._grid_floor(y, 1)

    def grid_ceil_x(self, point):
        x, y = point
        return self._grid_ceil(x, 0), y

    def grid_ceil_y(self, point):
        x, y = point
        return x, self._grid_ceil(y, 1)

    def grid_ceil_xy(self, point):
        x, y = point
        return self._grid_ceil(x, 0), self._grid_ceil(y, 1)

    def grid_round_x(self, point):
        x, y = point
        return self._grid_round(x, 0), y

    def grid_round_y(self, point):
        x, y = point
        return x, self._grid_round(y, 1)

    def grid_round_xy(self, point):
        x, y = point
        return self._grid_round(x, 0), self._grid_round(y, 1)

    def neigborhood(self, point: Tuple[int, int], max_distance: int, norm_ord=2):
        """ Get all points on grid that are at most `max_distance` away from `point`.
        :param point: Reference point.
        :param max_distance:
        :param norm_ord: The order of the norm to be used. Default = 2 (euclidian norm)
        :return:
        """
        offset = self.grid_floor_xy(point)
        start = self.grid_floor_xy((offset[0] - max_distance, offset[1] - max_distance))
        end = self.grid_ceil_xy(
            (offset[0] + max_distance + self.step[0] + 1, offset[1] + max_distance + self.step[1] + 1))

        kernel = Grid2D(start, end, self.step)

        mesh = kernel.meshgrid()
        xx, yy = mesh
        x, y = point
        diff = xx - x, yy - y
        dists = np.linalg.norm(np.array(diff), ord=norm_ord, axis=0)

        neighbor_mask = dists <= max_distance

        neighbor_points = zip(*(x[neighbor_mask].flat for x in mesh))
        return [p for p in neighbor_points if self.is_on_grid(p)]


def test_grid2d():
    g = Grid2D((0, 0), (10, 20), (1, 2))

    assert g.is_on_grid((0, 0))
    assert not g.is_on_grid((1, 1))
    assert g.is_on_grid((1, 2))

    assert g.grid_floor_xy((1, 1)) == (1, 0)
    assert g.grid_ceil_xy((1, 1)) == (1, 2)


def test_grid_neighbors():
    for i in range(0, 1000):
        start = (np.random.randint(-10, 10), np.random.randint(-10, 10))
        end = (np.random.randint(10, 20), np.random.randint(10, 20))
        step = (np.random.randint(1, 4), np.random.randint(1, 4))
        g = Grid2D(start, end, step)
        point = np.random.random(2) * 10
        max_dist = np.random.random(1) * 10
        norm_ord = np.random.randint(1, 3)
        neighbors_expected = [p for p in g if np.linalg.norm(np.array(p) - np.array(point), ord=norm_ord) <= max_dist]
        neighbors_actual = g.neigborhood(point, max_dist, norm_ord=norm_ord)

        # print(set(neighbors_actual).symmetric_difference(set(neighbors_expected)))

        assert set(neighbors_actual) == set(neighbors_expected)



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from enum import Enum
from typing import Any, Dict, List, Tuple, Set, Union
from itertools import chain

"""
Data structure for generating LEF files.

References:
[LEF/DEF]: LEF/DEF 5.7 Language Reference
"""

FLOAT_FORMAT = "%.8f"


class LefFormattable:

    def format(self):
        return None


class LefStatement(LefFormattable):
    pass


class LefContainer(LefStatement):
    pass


class Direction(Enum):
    INPUT = 1
    OUTPUT = 2
    OUTPUT_TRISTATE = 3
    INOUT = 4
    FEEDTHRU = 5


class Use(Enum):
    SIGNAL = 0
    ANALOG = 1
    POWER = 2
    GROUND = 3
    CLOCK = 4


class Shape(Enum):
    ABUTMENT = 0
    RING = 1
    FEEDTHRU = 2


class Symmetry(LefStatement, Enum):
    X = 1
    Y = 2
    R90 = 4

    def __hash__(self):
        return self.value


class Class(Enum):
    NONE = 0
    CORE = 1
    BUMP = 2


class MacroClass(LefStatement, Enum):
    COVER = 10
    COVER_BUMP = 11
    RING = 20
    BLOCK = 30
    BLOCK_BLACKBOX = 31
    BLOCK_SOFT = 32
    PAD = 40
    PAD_INPUT = 41
    PAD_OUTPUT = 42
    PAD_INOUT = 43
    PAD_POWER = 44
    PAD_SPACER = 45
    PAD_AREAIO = 46
    CORE = 50
    CORE_FEEDTHRU = 51
    CORE_TIEHIGH = 52
    CORE_TIELOW = 53
    CORE_SPACER = 54
    CORE_ANTENNACELL = 55
    CORE_WELLTAP = 56
    ENDCAP_PRE = 61
    ENDCAP_POST = 62
    ENDCAP_TOPLEFT = 63
    ENDCAP_TOPRIGHT = 64
    ENDCAP_BOTTOMLEFT = 65
    ENDCAP_BOTTOMRIGHT = 66

    def format(self):
        return "CLASS {}".format(self.name.replace("_", " ").upper())


class Point(LefFormattable):

    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def format(self):
        return "{} {}".format(self.x, self.y)


class Property(LefStatement):

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def format(self):
        return "{} {}".format(self.name.upper(), self.value)


class Foreign(LefStatement):
    """ Reference to the GDSII data structure and its offset relative to the macro.
    """

    def __init__(self, name, offset: Point):
        self.name = name
        self.offset = offset

    def format(self):
        return "{} {}".format(self.name.upper(), self.offset.format())


class Polygon(LefStatement):

    def __init__(self, points: List[Tuple[int, int]]):
        self._points = points

    def format(self) -> str:
        return "POLYGON {}".format(" ".join(
            (FLOAT_FORMAT % x for x in chain(*self._points)))
        )


class Rect(LefStatement):

    def __init__(self, point1: Tuple[int, int], point2: Tuple[int, int]):
        self._points = [point1, point2]

    def format(self) -> str:
        return "RECT {}".format(" ".join(
            (FLOAT_FORMAT % x for x in chain(*self._points)))
        )


class Layer(LefStatement):

    def __init__(self, layer_name: str):
        self.name = layer_name

    def format(self) -> str:
        return "LAYER {}".format(self.name)


class Port(LefContainer):

    def __init__(self, CLASS: Class, geometries: List[Tuple[Layer, List[Union[Rect, Polygon]]]]):
        self.CLASS = CLASS  # Class
        self.geometries = geometries

    def format(self):
        return [
            "PORT",
            [
                self.CLASS,
                [[l, g] for l, g in self.geometries]
            ],
            "END"
        ]


class Pin(LefContainer):
    """
    [LEF/DEF] p. 193.
    """

    def __init__(self, pin_name: str, direction: Direction, use: Use, shape: Shape, port: Port,
                 property: Dict[str, Any]):
        self.name = pin_name
        self.direction = direction
        self.use = use
        self.shape = shape
        self.port = port
        self.property = property  # {property_name: value, ...}

    def format(self):
        return [
            "PIN {}".format(self.name),
            [
                self.direction,
                self.use,
                self.shape,
                self.port.format(),
            ],
            "END {}".format(self.name),
            []  # Empty line.
        ]


class Obstruction(LefContainer):
    """
    Blockage
    """

    def __init__(self, layer: Layer, geometries: List):
        self.layer = layer
        self.geometries = geometries

    def format(self):
        return [
            "OBS",
            [
                self.layer,
                [g for g in self.geometries]
            ],
            "END",
            []
        ]


class Macro(LefContainer):
    """
    [LEF/DEF] p. 172
    """

    def __init__(self,
                 name: str,
                 macro_class: MacroClass,
                 foreign: Foreign,
                 obstructions: List[Obstruction],
                 origin: Point,
                 symmetry: Set[Symmetry],
                 pins: List[Pin],
                 site: str,
                 property: Dict[str, Any] = None,
                 ):
        self.name = name
        self.obstructions = obstructions  # Obstructions (Blockages), [LEF/DEF] p. 192.
        self.macro_class = macro_class
        self.foreign = foreign
        self.origin = origin
        self.symmetry = symmetry
        self.pins = pins
        self.property = property if property is not None else {}
        self.site = site

    def format(self):
        return [
            "MACRO {}".format(self.name.upper()),
            [
                self.macro_class,
                Property("ORIGIN", self.origin.format()),
                Property("FOREIGN", self.foreign.format()),
                Property("SITE", self.site.upper()),
                Property("SYMMETRY", " ".join((s.name.upper() for s in self.symmetry)))
            ],
            self.pins,
            self.obstructions,
            "END {}".format(self.name.upper()),
            [],
        ]


class LibraryLEF(LefContainer):
    def __init__(self,
                 version: str,
                 busbitchars: str = '[]',
                 dividerchar: str = '/',
                 macros: List[Macro] = None
                 ):
        self.version = version
        self.busbitchars = busbitchars
        self.dividerchar = dividerchar
        self.macros = macros if macros is not None else []

    def format(self):
        return [
            Property("VERSION", self.version),
            Property("NAMESCASESENSITIVE", "ON"),
            Property("BUSBITCHARS", '"{}"'.format(self.busbitchars)),
            Property("DIVIDERCHAR", '"{}"'.format(self.dividerchar)),
            self.macros,
            "END LIBRARY"
        ]


def lef_format(lef_obj, indent=-1, indent_char=' ') -> str:
    """ Convert LEF data structure into a string.
    :param lef_obj:
    :param indent:
    :param indent_char:
    :return: LEF formatted as a string.
    """
    i = indent_char * max(0, indent)
    if isinstance(lef_obj, LefContainer):
        return lef_format(lef_obj.format(), indent, indent_char)
    elif isinstance(lef_obj, list):
        return "\n".join((lef_format(e, indent + 1, indent_char) for e in lef_obj))
    elif isinstance(lef_obj, LefStatement):
        return i + "{} ;".format(lef_obj.format())
    elif isinstance(lef_obj, str):
        return i + lef_obj
    elif isinstance(lef_obj, Enum):
        return i + "{} {} ;".format(type(lef_obj).__name__.upper(), lef_obj.name.upper())
    else:
        assert False, "Unsupported type: {} ({})".format(type(lef_obj), lef_obj)


def test_lef():
    port = Port(Class.CORE,
                [(Layer('metal1'),
                  [Polygon([(0, 0), (1, 1)]), Polygon([(0, 0), (1, 1)]),
                   Rect((1, 2), (3, 4)), Rect((1, 2), (3, 4))])])

    pin = Pin(pin_name='pinName',
              direction=Direction.INPUT,
              use=Use.SIGNAL,
              shape=Shape.ABUTMENT,
              property={},
              port=port)

    macro = Macro('AND2X8',
                  macro_class=MacroClass.CORE,
                  foreign=Foreign('foreign_name', Point(0, 0)),
                  obstructions=[Obstruction(Layer('metal1'), [Rect((1, 2), (3, 4))])],
                  origin=Point(0, 0),
                  pins=[pin, pin],
                  property={},
                  symmetry={Symmetry.X, Symmetry.Y, Symmetry.R90},
                  site='someSite')

    library = LibraryLEF(version="0.0", macros=[macro])

    s = lef_format(library)

    print(s)


"""
Order of LEF Statements
[VERSION statement]
[BUSBITCHARS statement]
[DIVIDERCHAR statement]
[UNITS statement]
[MANUFACTURINGGRID statement]
[USEMINSPACING statement]
[CLEARANCEMEASURE statement ;]
[PROPERTYDEFINITIONS statement]
[ LAYER (Nonrouting) statement
| LAYER (Routing) statement] ...
[MAXVIASTACK statement]
[VIA statement] ...
#Fixed vias that can be used inside VIARULE
[VIARULE statement] ...
[VIARULE GENERATE statement] ...
[VIA statement] ...
#Generated vias that can reference VIARULE name
[NONDEFAULTRULE statement] ...
[SITE statement] ...
[MACRO statement
[PIN statement] ...
[OBS statement ...]] ...
[BEGINEXT statement] ...
[END LIBRARY]
"""

"""
Library LEF file
[VERSION statement]
[BUSBITCHARS statement]
[DIVIDERCHAR statement]
[VIA statement] ...
[SITE statement]
[MACRO statement
[PIN statement] ...
[OBS statement ...] ] ...
[BEGINEXT statement] ...
[END LIBRARY]
"""



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from .place import TransistorPlacer
from ..extrema import all_min, all_max
from lccommon import net_util

from ..data_types import *

import numpy as np
from itertools import tee, chain, cycle, islice, permutations, product
from typing import Iterable, Tuple, Hashable, List, Set, Any, Dict

import networkx as nx

from . import eulertours
from . import partition

import logging

logger = logging.getLogger(__name__)


def window(iterable, size):
    iters = tee(iterable, size)
    for i in range(1, size):
        for each in iters[i:]:
            next(each, None)
    return zip(*iters)


def _trim_none(l: List) -> List:
    """ Cut away all `None`s from the start and end of the list `l`.
    :param l:
    :return:
    """
    while l and l[0] is None:
        l = l[1:]
    while l and l[-1] is None:
        l = l[:-1]
    return l


def _remove_gaps(cell: Cell) -> Cell:
    """
    Remove unnecessary diffusion gaps where possible.
    :param cell: The input cell.
    :return: The compacted.
    """

    upper = []
    lower = []

    n = len(cell.upper)
    assert n == len(cell.lower)

    for i in range(n):
        l = cell.lower[i]
        u = cell.upper[i]

        if u is not None or l is not None:
            upper.append(u)
            lower.append(l)
        else:
            # Both places are empty.
            assert u is None and l is None

            # Place the gap only if it is need to avoid a short circuit.
            if i >= 1 and i + 1 < n:
                prev_u = upper[-1]
                prev_l = lower[-1]
                next_u = cell.upper[i + 1]
                next_l = cell.lower[i + 1]

                # Check if the upper or lower row needs the diffusion gap.
                need_gap = __need_gap(prev_u, next_u) or __need_gap(prev_l, next_l)

                if need_gap:
                    upper.append(None)
                    lower.append(None)

    return _assemble_cell(lower, upper)


def _assemble_cell(lower_row: List[Transistor], upper_row: List[Transistor]) -> Cell:
    """ Build a Cell object from a nmos and pmos row.
    :param lower_row:
    :param upper_row:
    :return:
    """
    width = max(len(lower_row), len(upper_row))
    cell = Cell(width)
    for i, t in enumerate(upper_row):
        cell.upper[i] = t

    for i, t in enumerate(lower_row):
        cell.lower[i] = t
    return cell


def _row_io_ordering_cost(row: List[Transistor], input_nets: Set[Any], output_nets: Set[Any]) -> float:
    """ Return a low cost if input nets are placed on the left side and outputs on the right side.
    """

    nets = [(net, pos) for pos, net in
            enumerate(chain(*[[t.source_net, t.gate_net, t.drain_net] for t in row if t is not None]))
            ]
    return _io_ordering_cost(nets, input_nets, output_nets)


def _io_ordering_cost(nets: Iterable[Tuple[Hashable, float]], input_nets: Set[Any], output_nets: Set[Any]) -> float:
    """ Return a low cost if input nets are placed on the left side and outputs on the right side.
    :param nets: Tuples of (net name, x-position)
    :param input_nets: Names of input nets.
    :param output_nets: Names of output nets.
    :return: Returns a cost which prefers that input nets are placed on the left and output nets are placed on the right.
    """

    nets = list(nets)

    cost = 0
    min_x = min((pos for _, pos in nets))
    max_x = max((pos for _, pos in nets))

    for net, pos in nets:
        if net in input_nets:
            cost += pos - min_x
        elif net in output_nets:
            cost += max_x - pos

    return cost


def _num_gate_matches(cell: Cell) -> int:
    """
    Count how many opposing gates share the same net.
    :param cell:
    :return: Number of gate matches.
    """
    sum = 0
    for a, b in zip(cell.upper, cell.lower):
        if a is not None and b is not None:
            if a.gate_net == b.gate_net:
                sum += 1
    return sum


def _cell_quality(cell: Cell, input_nets, output_nets):
    """ Quality metric for a dual stack cell.
    :param nmos:
    :param pmos:
    :param input_nets:
    :param output_nets:
    :return:
    """
    return _num_gate_matches(cell), -wiring_length_bbox(cell), \
           - _row_io_ordering_cost(cell.upper, input_nets, output_nets) \
           - _row_io_ordering_cost(cell.lower, input_nets, output_nets)


def __need_gap(left: Transistor, right: Transistor):
    """Check if a diffusion gap is required between two adjacent transistors.
    """
    if left is None or right is None:
        return False
    else:
        return left.drain_net != right.source_net


def _wiring_length_bbox1(nets: Iterable[Tuple[Hashable, float]]) -> float:
    """ Calculate 1-dimensional wiring length.
    :param nets: List[(net name, x)]
    :return: Sum of bounding box circumferences
    """

    minmax_x = dict()

    for net, x in nets:
        minmax_x[net] = (
            min(x, minmax_x.get(net, (x, x))[0]),
            max(x, minmax_x.get(net, (x, x))[1])
        )

    bbox_width = {k: mx - mn for k, (mn, mx) in minmax_x.items()}
    return sum(bbox_width.values())


def test_wiring_length_bbox1():
    nets = {('a', 0), ('b', 2), ('a', 1), ('b', 12)}
    assert _wiring_length_bbox1(nets) == 11


def _wiring_length_bbox2(nets: Iterable[Tuple[Hashable, Tuple[float, float]]]) -> float:
    """ Calculate 2-dimensional wiring length.
    :param nets: List[(net name, (x,y))]
    :return: Sum of bounding box circumferences
    """

    len_x = _wiring_length_bbox1((net, x) for net, (x, y) in nets)
    len_y = _wiring_length_bbox1((net, y) for net, (x, y) in nets)

    return len_x + len_y


def wiring_length_bbox(cell: Cell) -> float:
    """Get an indicator of the wiring length inside a cell.
    The bounding box size of the nets is used as an approximation.
    """

    net_positions = []
    for row in (cell.upper, cell.lower):
        for pos, net in enumerate(chain(*[list(t.terminals()) for t in row if t is not None])):
            net_positions.append((net, pos))

    return _wiring_length_bbox1(net_positions)


class HierarchicalPlacer(TransistorPlacer):

    def __init__(self):
        pass

    def place(self, transistors: Iterable[Transistor]) -> Cell:
        """ Place transistors by a hierarchical approach.
        The full circuit is split into sub-circuits, each containing only nmos or pmos transistors.
        The sub-circuits are placed independent of their internal placement.
        :param transistors:
        :return:
        """

        # Split into nmos / pmos
        nmos = [t for t in transistors if t.channel_type == ChannelType.NMOS]
        pmos = [t for t in transistors if t.channel_type == ChannelType.PMOS]

        nmos_graph = _transistors2graph(nmos)
        pmos_graph = _transistors2graph(pmos)

        # Partition into sub-circuits.
        nmos_graphs = partition.partition(nmos_graph)
        pmos_graphs = partition.partition(pmos_graph)

        def find_internal_placements(g: nx.MultiGraph) -> List[List[Transistor]]:
            """Find optimal-width placements for each sub-circuit.
            """
            assert len(g) > 0
            placements = _find_optimal_single_row_placements(g)
            # Append diffusion gap.
            placements = [list(chain(p, [None])) for p in placements]
            return placements

        # Find optimal-width placements for each sub-circuit.
        nmos_placements = [find_internal_placements(g) for g in nmos_graphs]
        pmos_placements = [find_internal_placements(g) for g in pmos_graphs]

        class SubCell:
            """ Abstract representation of a set of transistors.
            Only the width of the sub cell and a set of terminal names is stored.
            """

            def __init__(self, width, nets, row, id):
                """

                :param width: Width of sub cell
                :param nets: Set of net names inside the sub-cell (without power nets)
                :param row: Row index as a placement constraint. (e.g. 0 for a NMOS sub cell and 1 for a PMOS sub cell)
                :param id: Some identifier or reference to the transistors in this sub cell.
                """
                self.width = width
                self.nets = nets
                self.row = row
                self.id = id
                self.location = (None, None)

            def __repr__(self):
                return "SubCell(%d, %s)" % (self.width, self.nets)

        def subcell_from_placements(placements: List[List[Transistor]], row: int) -> SubCell:
            """ Create a SubCell from a list of possible internal placements.
            :param placements:
            :param row:
            :return:
            """
            if len(placements) == 0:
                return SubCell(0, set(), 0, placements)

            transistors = [t for t in chain(*placements) if t is not None]
            # Extract net names
            nets = {n for n in chain(*(t.terminals() for t in transistors))}
            # Strip power nets
            nets = {n for n in nets if not net_util.is_power_net(n)}

            return SubCell(len(placements[0]), nets, row, placements)

        # Get sub cells
        p_subcells = [subcell_from_placements(pl, 0) for pl in pmos_placements]
        n_subcells = [subcell_from_placements(pl, 1) for pl in nmos_placements]

        p_permuations = list(permutations(p_subcells))
        n_permuations = list(permutations(n_subcells))

        def get_subcell_net_position(subcells: Iterable[SubCell]) -> List[Tuple[Any, float]]:
            """
            Get approximate net positions given a row of sub-cells.
            Each net is treated as it would be localized in the center of the sub cell.
            :param subcells: A row of sub-cells
            :return: List[(net name, x coordinate)]
            """
            offset = 0
            nets = []
            for subcell in subcells:
                x = offset + subcell.width / 2
                nets.extend(((net, x) for net in subcell.nets))
                offset += subcell.width
            return nets

        # Get optimal placement of sub cells by trying all combinations.
        # Quality metric is the bounding box size of the nets.
        subcell_placements = product(p_permuations, n_permuations)

        # Get good sub cell placements by evaluating the bounding box sizes of the nets.
        best_subcell_placements = []
        best_wiring_length = None

        # Extract input nets.
        input_nets = net_util.get_cell_inputs(transistors)
        output_nets = set()  # TODO: extract output nets.

        for ps, ns in subcell_placements:
            ppos = get_subcell_net_position(ps)
            npos = get_subcell_net_position(ns)

            # TODO: make quality metric parametrizable.
            # Optimize for wiring length estimate and break ties with the ordering of IO nets.
            # Want input nets to be on the left.
            wiring_length = _wiring_length_bbox1(chain(ppos, npos))
            cost = wiring_length, _io_ordering_cost(chain(ppos, npos), input_nets, output_nets)

            if best_wiring_length is None or cost < best_wiring_length:
                best_subcell_placements.clear()
                best_wiring_length = cost
            if cost <= best_wiring_length:
                best_subcell_placements.append((ps, ns))

        logger.debug('Number of best sub cell placements: %d', len(best_subcell_placements))

        def get_subcell_x_offsets(subcell_row: Iterable[SubCell], offset: int = 0) -> Dict[Any, float]:
            """ Given a row of subcells find the x offset of each.
            """
            offsets = dict()
            for subcell in subcell_row:
                offsets[subcell] = offset
                offset += subcell.width
            return offsets

        def find_best_intra_subcell_placement(ps: List[SubCell], ns: List[SubCell]) -> Dict[SubCell, List[Transistor]]:
            """ Given the placement of the sub cells find the best transistor placements
            inside the subcells.
            :param ps: PMOS sub cells
            :param ns: NMOS sub cells
            :return:
            """

            # Get full cell width
            pwidth = sum((sc.width for sc in ps))
            nwidth = sum((sc.width for sc in ns))
            width = max(pwidth, nwidth)

            # Get a map of which track is occupied by which sub cell.
            height = 2
            matrix = np.ndarray((height, width), dtype=object)
            for row_idx, subcells in enumerate([ns, ps]):
                row = list(chain(*([s] * s.width for s in subcells)))
                # Pad to width
                row.extend([None] * (width - len(row)))
                matrix[row_idx, :] = row

            # Build a dependency graph: Find sub cells that depend on each others internal placement
            # (e.g. sub cells that face each other)
            dependency_graph = nx.Graph()
            # Add all nodes to graph
            for sc in chain(ns, ps):
                dependency_graph.add_node(sc)
            # Check if neighbors (in y directon) share a net. If yes, add an edge in the dependency graph.
            # Loop over pairs of neighbors in y-direction
            for a, b in zip(matrix[0:].flat, matrix[1:].flat):
                if a is not None and b is not None:
                    common_nets = a.nets & b.nets
                    if len(common_nets) > 0:
                        dependency_graph.add_edge(a, b)

            # Split dependency graph into connected sub graphs.
            subgraphs = [dependency_graph.subgraph(c) for c in nx.connected_components(dependency_graph)]
            # Construct groups of dependent cells.
            # The internal placement between different groups is independent.
            dependent_cell_groups = [list(g.nodes) for g in subgraphs]

            logger.debug('Number of sub cells: %d', len(dependent_cell_groups))

            # Find x positions of sub cells (left corner).
            x_offsets = dict()
            x_offsets.update(get_subcell_x_offsets(ps))
            x_offsets.update(get_subcell_x_offsets(ns))

            best_intra_subcell_placements = dict()

            # Iterate over groups of sub cells that must be placed together.
            for subcells in dependent_cell_groups:

                subcell_net_positions = set()
                # Add simplified net positions of all other sub cells.
                for sc in chain(ps, ns):
                    if sc not in subcells:
                        p = x_offsets[sc] + sc.width / 2
                        for net in sc.nets:
                            subcell_net_positions.add((net, p))

                min_wiring_length = None
                best_placements = []
                # Iterate over all internal placements

                for placements in product(*(s.id for s in subcells)):
                    # Get net positions for this placement
                    assert len(subcells) == len(placements)
                    net_positions = []
                    for subcell, placement in zip(subcells, placements):

                        # Expand placement into transistor terminals
                        for net_pos, net in enumerate(chain(*(t.terminals() for t in placement if t))):
                            # Ignore power nets
                            if not net_util.is_power_net(net):
                                p = x_offsets[subcell] + net_pos / 3
                                net_positions.append((net, p))

                    # Add simplified net positions of all other sub cells.
                    net_positions.extend(subcell_net_positions)

                    wiring_length = _wiring_length_bbox1(net_positions)

                    if min_wiring_length is None or wiring_length < min_wiring_length:
                        min_wiring_length = wiring_length
                        best_placements.clear()

                    if min_wiring_length == wiring_length:
                        best_placements.append(placements)

                best_placement = best_placements[0]
                assert len(subcells) == len(best_placement)
                for subcell, placement in zip(subcells, best_placement):
                    best_intra_subcell_placements[subcell] = placement

            return best_intra_subcell_placements

        # Find best internal placement for each cell in `subcells`

        best_quality = None
        best_cell = None

        for ps, ns in best_subcell_placements:
            intra_placements = find_best_intra_subcell_placement(ps, ns)

            p_row = []
            n_row = []

            for p in ps:
                transistors = intra_placements[p]
                p_row.extend(transistors)
            for n in ns:
                transistors = intra_placements[n]
                n_row.extend(transistors)

            cell = _assemble_cell(n_row, p_row)

            # Remove gaps where possible.
            cell = _remove_gaps(cell)

            input_nets = net_util.get_cell_inputs(transistors)
            output_nets = {}
            quality = _cell_quality(cell, input_nets, output_nets)

            if best_quality is None or quality > best_quality:
                best_quality = quality

            if quality == best_quality:
                best_cell = cell

        return best_cell


def _transistors2graph(transistors: Iterable[Transistor]) -> nx.MultiGraph:
    """ Create a graph representing the transistor network.
        Each edge corresponds to a transistor, each node to an electrical potential.
    """
    G = nx.MultiGraph()
    for t in transistors:
        G.add_edge(t.source_net, t.drain_net, t)
    # assert nx.is_connected(G)
    return G


def _find_optimal_single_row_placements(transistor_graph: nx.MultiGraph) -> List[List[Transistor]]:
    """ Find with-optimal single row placements of transistors.

    :param transistors: nx.MultiGraph representing the transistor network. Each edge coresponds to a transistor.
    :return: List[List[Transistor]]
    """

    even_degree_graphs = eulertours.construct_even_degree_graphs(transistor_graph)
    logger.debug('Number of even-degree graphs: %d', len(even_degree_graphs))

    all_eulertours = list(chain(*(eulertours.find_all_euler_tours(g) for g in even_degree_graphs)))

    logger.debug('Number of eulertours: %d', len(all_eulertours))
    all_eulertours = list(set((tuple(tour) for tour in all_eulertours)))
    logger.debug('Number of deduplicated eulertours: %d', len(all_eulertours))

    def cyclic_shift(l: List, shift: int) -> List:
        """ Rotate elements of a list `l` by `shift` to the left.

        :param l: List
        :param shift: Shift amount.
        :return: A shifted copy of the list.
        """
        s = islice(cycle(l), shift, len(l) + shift)
        return list(s)

    def find_optimal_shifts(tours: List[List[Transistor]]) -> List[List[Transistor]]:
        """ Find the optimal euler tours by applying a cyclic rotation to all provided euler tours.
        """
        if not tours:
            return []

        optimal = []
        optimal_len = len(tours[0])

        for tour in tours:
            tour = _trim_none(tour)
            assert len(tour) >= 1

            if tour[0].source_net != tour[-1].drain_net:
                # Can not be shifted without inserting a gap.
                tour.append(None)

            l = len(tour)
            for i in range(l):
                shifted = cyclic_shift(tour, i)
                if shifted[0] is not None:
                    shifted = _trim_none(shifted)
                    l = len(shifted)

                    if l < optimal_len:
                        optimal.clear()
                        optimal_len = l
                    if l == optimal_len:
                        optimal.append(shifted)

        # Check if there are duplicates
        _optimal = (tuple(t) for t in optimal)
        _optimal_set = set(_optimal)
        # assert len(_optimal) == len(_optimal_set)
        # Deduplicate
        optimal = list(_optimal_set)

        return optimal

    def edges2transistors(edges: Tuple[Any, Any, int]) -> List[Transistor]:
        """ Convert graph edges back to a list of transistors.
        """

        ts = []
        for l, r, t in edges:
            transistor = None
            if isinstance(t, Transistor):
                if t.source_net == l:
                    transistor = t
                else:
                    transistor = t.flipped()

                assert transistor.source_net == l and transistor.drain_net == r

            ts.append(transistor)
        return ts

    assert len(all_eulertours) > 0
    all_placements = [edges2transistors(edges) for edges in all_eulertours]
    assert len(all_placements) > 0
    all_placements = find_optimal_shifts(all_placements)

    assert len(all_placements) > 0
    return all_placements


class EulerPlacer(TransistorPlacer):

    def __init__(self):
        pass

    def place(self, transistors: Iterable[Transistor]) -> Cell:
        """Find a 2-stack placement for the tansistors.

        Transistors will be placed on a 2xn grid (2-stack cell). PMOS on the upper stack, NMOS on the lower stack.
        The placement locations are not absolute coordinates but two dimensional indices to the transistor grid.
        """
        transistors = list(transistors)
        logger.debug('Find eulerian tours.')

        nmos = [t for t in transistors if t.channel_type == ChannelType.NMOS]
        pmos = [t for t in transistors if t.channel_type == ChannelType.PMOS]

        nmos_graph = _transistors2graph(nmos)
        pmos_graph = _transistors2graph(pmos)

        all_nmos = _find_optimal_single_row_placements(nmos_graph)
        all_pmos = _find_optimal_single_row_placements(pmos_graph)

        logger.debug('Number of NMOS placements with cyclic shifts: %d', len(all_nmos))
        logger.debug('Number of PMOS placements with cyclic shifts: %d', len(all_pmos))

        if len(all_nmos) * len(all_pmos) > 100000:
            # Notify the user that the hierarchical placer might be better.
            logger.info('`EulerPlacer` will not perform well in this case. '
                        '`HierarchicalPlacer` could be a better choice.')

        # Find best nmos/pmos row pair.
        pairs = product(all_nmos, all_pmos)

        input_nets = net_util.get_cell_inputs(transistors)
        output_nets = {}

        # Assemble optimal cell candidates.
        cells = (_assemble_cell(nmos, pmos) for nmos, pmos in pairs)

        best_cells_gate_match = all_max(cells, key=_num_gate_matches)
        best_cells_wiring = all_min(best_cells_gate_match, key=wiring_length_bbox)

        def io_ordering_cost(cell: Cell):
            return _row_io_ordering_cost(cell.lower, input_nets, output_nets) + \
                   _row_io_ordering_cost(cell.upper, input_nets, output_nets)

        best_cells_io_ordering = all_min(best_cells_wiring, key=io_ordering_cost)
        best_cells = best_cells_io_ordering

        if len(best_cells) > 1:
            logger.info("Found multiple optimal placements: %d. Take the first.", len(best_cells))

        return best_cells[0]



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from klayout import db
from ..layout.layers import *
from typing import Dict, List
import logging

logger = logging.getLogger(__name__)


class MOS4To3NetlistSpiceReader(db.NetlistSpiceReaderDelegate):
    """
    Read SPICE netlists and convert 4-terminal MOS into 3-terminal MOS by dropping the body net.
    This is required for the LVS step when the standard cells are lacking well taps and therefore
    the body terminal of the transistors is unconnected.
    """

    def element(self, circuit: db.Circuit, el: str, name: str, model: str, value, nets: List[db.Net],
                params: Dict[str, float]):
        """
        Process a SPICE element. All elements except 4-terminal MOS transistors are left unchanged.
        :return: True iff the device has not been ignored and put into the netlist.
        """
        if el != 'M' or len(nets) != 4:
            # All other elements are left to the standard implementation.
            return super().element(circuit, el, name, model, value, nets, params)
        else:
            # Provide a device class.
            cls = circuit.netlist().device_class_by_name(model)
            if not cls:
                # Create MOS3Transistor device class if it does not yet exist.
                cls = db.DeviceClassMOS3Transistor()
                cls.name = model
                circuit.netlist().add(cls)

            # Create MOS3 device.
            device: db.Device = circuit.create_device(cls, name)
            # Configure the MOS3 device.
            for terminal_name, net in zip(['S', 'G', 'D'], nets):
                device.connect_terminal(terminal_name, net)

            # Parameters in the model are given in micrometer units, so
            # we need to translate the parameter values from SI to um values.
            device.set_parameter('W', params.get('W', 0) * 1e6)
            device.set_parameter('L', params.get('L', 0) * 1e6)

            return True


def extract_netlist(layout: db.Layout, top_cell: db.Cell) -> db.Netlist:
    """
    Extract a device level netlist of 3-terminal MOSFETs from the cell `top_cell` of layout `layout`.
    :param layout: Layout object.
    :param top_cell: The top cell of the circuit.
    :return: Netlist as a `klayout.db.Circuit` object.
    """

    # Without netlist comparision capabilities.
    l2n = db.LayoutToNetlist(db.RecursiveShapeIterator(layout, top_cell, []))

    def make_layer(layer_name: str):
        return l2n.make_layer(layout.layer(*layermap[layer_name]), layer_name)

    rnwell = make_layer(l_nwell)
    rpwell = make_layer(l_pwell)
    rndiff = make_layer(l_ndiffusion)
    rpdiff = make_layer(l_pdiffusion)
    rpoly = make_layer(l_poly)
    # rpoly_lbl = make_layer(l_poly_label)
    rdiff_cont = make_layer(l_diff_contact)
    rpoly_cont = make_layer(l_poly_contact)
    rmetal1 = make_layer(l_metal1)
    rmetal1_lbl = make_layer(l_metal1_label)
    rvia1 = make_layer(l_via1)
    rmetal2 = make_layer(l_metal2)
    rmetal2_lbl = make_layer(l_metal2_label)

    rpactive = rpdiff & rnwell
    rpgate = rpactive & rpoly
    rpsd = rpactive - rpgate

    rnactive = rndiff - rnwell
    rngate = rnactive & rpoly
    rnsd = rnactive - rngate

    l2n.register(rpactive, 'pactive')
    l2n.register(rpgate, 'pgate')
    l2n.register(rpsd, 'psd')

    l2n.register(rnactive, 'nactive')
    l2n.register(rngate, 'ngate')
    l2n.register(rnsd, 'nsd')

    # 3 terminal PMOS transistor device extraction
    pmos_ex = db.DeviceExtractorMOS3Transistor("PMOS")
    l2n.extract_devices(pmos_ex, {"SD": rpsd, "G": rpgate, "W": rnwell, "tS": rpsd, "tD": rpsd, "tG": rpoly})

    # 3 terminal NMOS transistor device extraction
    nmos_ex = db.DeviceExtractorMOS3Transistor("NMOS")
    l2n.extract_devices(nmos_ex, {"SD": rnsd, "G": rngate, "W": rpwell, "tS": rnsd, "tD": rnsd, "tG": rpoly})

    # # 4 terminal PMOS transistor device extraction
    # pmos_ex = db.DeviceExtractorMOS4Transistor("PMOS")
    # l2n.extract_devices(pmos_ex, {"SD": rpsd, "G": rpgate, "W": rnwell, "tS": rpsd, "tD": rpsd, "tG": rpoly, "tB": rnwell})
    #
    # # 4 terminal NMOS transistor device extraction
    # nmos_ex = db.DeviceExtractorMOS4Transistor("NMOS")
    # l2n.extract_devices(nmos_ex, {"SD": rnsd, "G": rngate, "W": rpwell, "tS": rnsd, "tD": rnsd, "tG": rpoly, "tB": rpwell})

    # Define connectivity for netlist extraction

    # Intra-layer
    l2n.connect(rvia1)
    l2n.connect(rpsd)
    l2n.connect(rnsd)
    l2n.connect(rpoly)
    l2n.connect(rdiff_cont)
    l2n.connect(rpoly_cont)
    l2n.connect(rmetal1)
    l2n.connect(rmetal2)
    # TODO: what if more than 2 metal layers?

    # Inter-layer
    l2n.connect(rpsd, rdiff_cont)
    l2n.connect(rnsd, rdiff_cont)
    l2n.connect(rpoly, rpoly_cont)
    l2n.connect(rpoly_cont, rmetal1)
    l2n.connect(rdiff_cont, rmetal1)
    l2n.connect(rmetal1, rvia1)
    l2n.connect(rvia1, rmetal2)
    # l2n.connect(rpoly, rpoly_lbl)  # attaches labels
    l2n.connect(rmetal1, rmetal1_lbl)  # attaches labels
    l2n.connect(rmetal2, rmetal2_lbl)  # attaches labels

    # l2n.connect_global(rnwell, 'NWELL') # VDD
    # l2n.connect_global(rpwell, 'PWELL') # GND

    # Perform netlist extraction
    logger.debug("Extracting netlist from layout")
    l2n.extract_netlist()

    netlist = l2n.netlist()
    netlist.make_top_level_pins()
    netlist.purge()
    netlist.combine_devices()
    netlist.purge_nets()
    # netlist.simplify()

    assert netlist.top_circuit_count() == 1, "A well formed netlist should have exactly one top circuit."

    return netlist.dup()


def compare_netlist(extracted: db.Netlist, reference: db.Netlist) -> bool:
    """
    Check if two netlists are equal.
    Both netlists must contain only the circuit of the cell.
    Note: It is not possible to copy a circuit from one netlist into another. This makes `simplify()` fail. Better just
    delete all non-used circuits from the netlist.
    :param extracted:
    :param reference:
    :return: Returns True iff the two netlists are equivalent.
    """
    assert extracted.top_circuit_count() == 1, "Expected to get exactly one top level circuit."
    assert reference.top_circuit_count() == 1, "Expected to get exactly one top level circuit."

    # Make sure that combined/fingered transistors are compared correctly.
    # Bring transistors into a unique representation.
    reference.simplify()
    extracted.simplify()

    cmp = db.NetlistComparer()
    compare_result = cmp.compare(extracted, reference)
    logger.debug("Netlist comparision result: {}".format(compare_result))

    if not compare_result:
        logger.warning("Netlists don't match (use --verbose to display the netlists).")

        # Print the both netlists.
        logger.debug(f'''LVS netlists

LVS extracted netlist:
{extracted}

LVS reference netlist:
{reference}
''')

    return compare_result


def read_netlist_mos4_to_mos3(netlist_path: str) -> db.Netlist:
    """
    Read a SPICE netlist and convert all MOS4 transistors to MOS3 transistors.
    :param netlist_path:
    :return:
    """
    logger.debug("Loading netlist (convert MOS4 to MOS3): {}".format(netlist_path))
    netlist = db.Netlist()
    netlist.read(netlist_path, db.NetlistSpiceReader(MOS4To3NetlistSpiceReader()))
    return netlist



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from ..layout import layers
from klayout import db
from typing import Dict, List, Tuple, Union

import logging

logger = logging.getLogger(__name__)


class Writer:

    def write_layout(self,
                     layout: db.Layout,
                     pin_geometries: Dict[str, List[Tuple[str, db.Shape]]],
                     top_cell: db.Cell,
                     output_dir: str,
                     ) -> None:
        pass


def remap_layers(layout: db.Layout, output_map: Dict[str, Union[Tuple[int, int], List[Tuple[int, int]]]]) -> db.Layout:
    """
    Rename layer to match the scheme defined in the technology file.
    :param layout:
    :param output_map: Output mapping from layer names to layer numbers.
    :return:
    """
    logger.debug("Remap layers.")
    layout2 = db.Layout()

    for top1 in layout.each_cell():
        top2 = layout2.create_cell(top1.name)
        layer_infos1 = layout.layer_infos()
        for layer_info in layer_infos1:

            src_layer = (layer_info.layer, layer_info.datatype)

            if src_layer not in layers.layermap_reverse:
                msg = "Layer {} not defined in `layermap_reverse`.".format(src_layer)
                logger.warning(msg)
                dest_layers = src_layer
            else:
                src_layer_name = layers.layermap_reverse[src_layer]

                if src_layer_name not in output_map:
                    msg = "Layer '{}' will not be written to the output. This might be alright though.". \
                        format(src_layer_name)
                    logger.warning(msg)
                    continue

                dest_layers = output_map[src_layer_name]

            if not isinstance(dest_layers, list):
                dest_layers = [dest_layers]

            src_idx = layout.layer(layer_info)
            for dest_layer in dest_layers:
                dest_idx = layout2.layer(*dest_layer)
                top2.shapes(dest_idx).insert(top1.shapes(src_idx))

    return layout2



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import networkx as nx

from typing import Any, Dict, List, AbstractSet, Optional


class GraphRouter:

    def route(self,
              graph: nx.Graph,
              signals: Dict[Any, List[Any]],
              reserved_nodes: Optional[Dict[Any, AbstractSet[Any]]] = None,
              node_conflict: Optional[Dict[Any, AbstractSet[Any]]] = None,
              equivalent_nodes: Optional[Dict[Any, AbstractSet[Any]]] = None,
              is_virtual_node_fn=None
              ) -> Dict[Any, nx.Graph]:
        """

        :param graph: Routing graph.
        :param signals: Mapping of signal names to terminal nodes in the graph.
        :param reserved_nodes: Mapping of signal names to graph nodes that are reserved for this signal.
        :param node_conflict: Mapping of a node to other nodes that can not be used for routing at the same time.
        :param equivalent_nodes: For each node a set of nodes that are physically equivalent.
        :param is_virtual_node_fn: Function that returns True iff the argument is a virtual node.
        :return: Returns a dict mapping signal names to routing trees.
        """
        pass



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from pysmt.shortcuts import *
from pysmt.fnode import FNode
from pysmt.environment import get_env

from copy import deepcopy
from itertools import tee, permutations, chain, count, islice, cycle, combinations
from typing import Iterable, List, Tuple
from enum import Enum

get_env().enable_infix_notation = True


def window(iterable: Iterable, size: int) -> Iterable[Tuple]:
    """
    Get a sliding window iterator.
    :param iterable:
    :param size: Window size.
    :return: Iterator returning sliding windows.
    """
    iters = tee(iterable, size)
    for i in range(1, size):
        for each in iters[i:]:
            next(each, None)
    return zip(*iters)


def argmin(l: Iterable, key=lambda x: x) -> int:
    """
    Get index (position) of the minimal element.
    :param l: Iterable containing the elements.
    :param key: A lambda function to be applied to the elements before.
    :return: Index to the minimal element.
    """
    m, i = min(zip(map(key, l), count()))
    return i


class Orientation(Enum):
    """ Represents the orientation of a rectilinear polygon edge.
    """

    RIGHT = 0
    UP = 1
    LEFT = 2
    DOWN = 3

    def is_horizontal(self):
        """
        Returns `True` iff the edge points either to the right or to the left.
        :return:
        """
        return self == Orientation.RIGHT or self == Orientation.LEFT

    def is_vertical(self):
        """
        Returns `True` iff the edge points either downwards or upwards.
        :return:
        """
        return self == Orientation.UP or self == Orientation.DOWN

    def angle(self, other):
        """ Return the angle between `other` and `self` in degrees.

        Returns: 0, 90, 180 or 270
        """
        return ((4 + other.value - self.value) % 4) * 90

    def __str__(self):
        arrow_char = {
            self.RIGHT: '→',
            self.UP: '↑',
            self.LEFT: '←',
            self.DOWN: '↓'
        }
        return arrow_char[self]


class OEdge:
    """ Representation of a rectilinear line.

    Parameters:
    offset: Distance to (0, 0): x offset for vertical lines, y offset for horizontal lines.
    orientation: 'r', 'u', 'l' or 'd'
    original_coords: Original start and end points.
    endpoints: Start and end point.
    interval: The interval spanned by the edge.
    """

    def __init__(self, offset, orientation: Orientation,
                 original_coords: Tuple[Tuple[int, int], Tuple[int, int]]):
        self.offset = offset  # Distance to (0, 0)
        self.orientation = orientation  # Orientaion of the edge.
        self.original_coords = original_coords
        self.endpoints = None
        self.interval = None  # Sorted endpoints such that start < end.
        self.attrs = dict()

    @staticmethod
    def from_points(start: Tuple[int, int], end: Tuple[int, int]):
        """ Create an edge from start and end point.
        """
        (x1, y1), (x2, y2) = start, end
        assert (x1 == x2) ^ (y1 == y2), Exception("Edge must be either horizontal or vertical.")

        if x1 == x2:
            offset = x1
            if y1 < y2:
                orientation = Orientation.UP
            else:
                orientation = Orientation.DOWN
        else:
            offset = y1
            if x1 < x2:
                orientation = Orientation.RIGHT
            else:
                orientation = Orientation.LEFT

        return OEdge(offset, orientation, (start, end))

    def length(self):
        """
        Calculates the length of the edge. This is only possible if the `interval` is set.
        :return:
        """
        assert self.interval is not None, Exception("Cannot calculate edge length. (`interval` not set)")
        return self.interval[1] - self.interval[0]

    def angle(self, other) -> int:
        """
        Calculate the relative angle to another edge.
        :param other:
        :return:
        """
        return self.orientation.angle(other.orientation)

    def is_horizontal(self) -> bool:
        return self.orientation.is_horizontal()

    def is_vertical(self) -> bool:
        return self.orientation.is_vertical()

    def __repr__(self):
        return "%s %s" % (self.offset, self.orientation)


def points2opolygon(points: List[Tuple[int, int]]) -> List[OEdge]:
    """ Convert an orthogonal polygon into a unique, overhead less representation.

    Returns
    -------
    Returns a list of OEdge. The list contains horizontal and vertical edges in a strictly alternating order.
    """
    n = len(points)

    if n == 0:
        return []

    assert n % 2 == 0, Exception("Cannot convert an odd number of points into a orthogonal polygon.")

    # Find index of smallest point sorted lexicographically by (y,x).
    lower_left_idx = argmin(points, key=lambda p: tuple(reversed(p)))
    prev_point = points[lower_left_idx]

    e1 = OEdge.from_points(prev_point, points[(lower_left_idx + 1) % n])
    assert e1.orientation == Orientation.RIGHT, Exception("Polgon is not oriented CCW.")

    edges = []
    horiz = True
    for i in range(n):
        idx = (i + lower_left_idx + 1) % n
        x1, y1 = prev_point
        x2, y2 = points[idx]

        e = prev_point, points[idx]

        assert x1 == x2 or y1 == y2, Exception("Edge is not orthogonal.")

        edge = OEdge.from_points(*e)

        if horiz:
            assert y1 == y2
            assert edge.is_horizontal()
        else:
            assert x1 == x2
            assert edge.is_vertical()

        edges.append(edge)
        horiz = not horiz
        prev_point = x2, y2

    return edges


class OPolygon:
    """ Polygon with orthogonal edges.
    Edges and points are defined counter clock-wise.
    """

    def __init__(self, points=None, edges=None):
        assert points is not None or edges is not None, Exception("Either points or edges must be given.")

        self.edges = None
        if points is not None:
            self.edges = points2opolygon(points)
        elif edges is not None:
            self.edges = edges

    def __repr__(self):
        return "OPolygon(%s)" % self.edges

    def horizontals(self) -> List[OEdge]:
        """
        :return: Returns a list of all horizontal edges.
        """
        return [e for e in self.edges if e.is_horizontal()]

    def verticals(self) -> List[OEdge]:
        """
        :return: Returns a list of all vertical edges.
        """
        return [e for e in self.edges if e.is_vertical()]

    def edges_by_orientation(self, orientation: Orientation) -> List[OEdge]:
        """
        Get a list of edges filtered by orientation.
        :param orientation:
        :return:
        """
        return [e for e in self.edges if e.orientation == orientation]

    def points(self) -> List:
        """ Get points representing the polygon.
        Returns
        -------
        [(x1, y1), ...]
        """
        return opolygon2points(self.edges)


def opolygon2points(opolygon: OPolygon) -> List:
    """ Get the points representing the opolygon.
    """
    n = len(opolygon)
    if n == 0:
        return []

    assert n >= 4, Exception("Not a valid orthogonal polygon.")
    assert n % 2 == 0, Exception("Not a valid orthogonal polygon.")
    horiz = True
    points = []
    for j in range(n):
        i = j - 1

        if horiz:
            p = opolygon[i].offset, opolygon[j].offset
            points.append(p)
        else:
            p = opolygon[j].offset, opolygon[i].offset
            points.append(p)

        horiz = not horiz

    return points


class SOPolygon:
    """ Symbolic orthogonal polygon.

    Similar to OPolygon, but edge offsets are now pySMT symbols.
    """

    def __init__(self, name: str, opolygon: OPolygon):
        self.original_opolygon = opolygon
        self.edges = deepcopy(opolygon.edges)

        # Replace edge offsets by pySMT symbols.
        for i, e in enumerate(self.edges):
            e.attrs['orig_offset'] = e.offset
            if 'fixed' in e.attrs:
                # TODO remove this
                e.offset = Int(e.offset)
            else:
                e.offset = Symbol("%s_%s" % (name, i), INT)

        # Add endpoint pointers to edges.
        n = len(self.edges)
        for j in range(n):
            i = j - 1
            k = (j + 1) % n

            # Get previous, current and next edge
            e_prev = self.edges[i]
            e = self.edges[j]
            e_next = self.edges[k]

            e.endpoints = (e_prev.offset, e_next.offset)

            # Make sure e.interval is sorted ascending.
            if e.orientation in [Orientation.RIGHT, Orientation.UP]:
                e.interval = (e_prev.offset, e_next.offset)
            else:
                e.interval = (e_next.offset, e_prev.offset)

        self.name = name

    def __repr__(self):
        return "SOPolygon(%s)" % self.edges

    def horizontals(self) -> List[OEdge]:
        return [e for e in self.edges if e.is_horizontal()]

    def verticals(self) -> List[OEdge]:
        return [e for e in self.edges if e.is_vertical()]

    def edges_by_orientation(self, orientation) -> List[OEdge]:
        return [e for e in self.edges if e.orientation == orientation]

    def points(self) -> List[Tuple[FNode, FNode]]:
        return opolygon2points(self.edges)

    def area(self) -> FNode:
        """
        Get a formula for the area of the symbolic polygon.
        Warning: This returns a non-linear formula which can lead to very slow solving.
        Some solvers will not even support this.
        :return: Formula for the area.
        """
        vs = self.verticals()

        summands = []
        for v in vs:
            start, end = v.endpoints
            a = v.offset * (end - start)
            summands.append(a)

        return sum(summands)

    def assert_preserved_edge_orientation(self) -> FNode:
        """ Assert that an edge keeps its original direction.

        E.g. an edge originally pointing to the right is not allowed to point to the left.

        Returns
        -------
        pysmt.fnode.FNode
        """
        return And(*[e.interval[0] <= e.interval[1] for e in self.edges])

    def assert_absolute_fixed(self) -> FNode:
        """ Get a constraint that fixes the polygon to its original shape and absolute location.
        """
        return And(*[Equals(e.offset, Int(e.attrs['orig_offset'])) for e in self.edges])

    def to_opolygon(self, model):
        """ Convert the symbolic polygon back to a constant polygon.

        Parameters
        ----------
        model: The solution of the SMT solver.
        """
        const_edges = []
        for e in self.edges:
            off = e.offset
            off2 = model.get_py_value(off) if off.is_constant() else model[off].constant_value()
            edge = OEdge(off2, e.orientation, e.original_coords)

            const_edges.append(edge)

        return OPolygon(edges=const_edges)


def preserve_absolute_edge_order(edges: List[OEdge], min_distance: int = 0) -> FNode:
    """ Get a pySMT formula which constrains the edges to preserve their order.
    """

    if len(edges) > 0:
        for e in edges:
            assert e.orientation.is_vertical() == edges[0].orientation.is_vertical(), Exception(
                "Cannot mix edge orientations.")

    # Sort by original offset
    s = sorted(edges, key=lambda e: e.attrs['orig_offset'])

    constraints = [(b.offset - a.offset) >= min_distance
                   for a, b in window(s, 2)
                   if a.offset != b.offset
                   ]

    return And(*constraints)


def edges_contain_polygon(a: OEdge, b: OEdge) -> bool:
    """
    Check if the inside of the polygon is between the edges.
    :param a:
    :param b:
    :return:
    """
    return a.offset < b.offset and a.orientation == Orientation.DOWN and b.orientation == Orientation.UP \
           or a.offset < b.offset and a.orientation == Orientation.RIGHT and b.orientation == Orientation.LEFT


def min_width(edges: List[OEdge],
              min_distance: int,
              min_distance_other_direction: int = None,
              notch_instead_width=False) -> FNode:
    """ Get a pySMT formula which constrains the shape to a minimum width.
    :param edges: List of either all horizontal or vertical edges of a polygon.
    """

    if min_distance_other_direction is None:
        min_distance_other_direction = min_distance

    if len(edges) > 0:
        for e in edges:
            assert e.orientation.is_vertical() == edges[0].orientation.is_vertical(), Exception(
                "Cannot mix edge orientations.")

    # Sort by original offset
    s = sorted(edges, key=lambda e: e.attrs['orig_offset'])

    def is_edge_pair_relevant(a: OEdge, b: OEdge):
        if notch_instead_width:
            return edges_contain_polygon(b, a)
        else:
            return edges_contain_polygon(a, b)

    # Create conditional constraints:
    # If the edge projections overlap, then the minimum width must be asserted.
    # Edges of length 0 can be ignored.
    # If `notch_instead_width` is set, then invert the selection of edges.
    constraints = [
        Implies(
            And(
                a.length() > 0,
                b.length() > 0,
                edge_projections_overlap(a, b, margin=min_distance_other_direction)
            ),
            (b.offset - a.offset) >= min_distance
        )
        for a, b in combinations(s, 2)
        if a.offset != b.offset and is_edge_pair_relevant(a, b)
    ]

    return And(*constraints)


def min_width_poly(poly: SOPolygon, min_distance: int) -> FNode:
    """
    Get a minimum width constraint for a symbolic polygon.
    :param polys:
    :param min_distance:
    :return:
    """
    return And(
        min_width(poly.horizontals(), min_distance=min_distance),
        min_width(poly.verticals(), min_distance=min_distance)
    )


def min_width_of_polygons(polys: Iterable[SOPolygon], min_width: int) -> FNode:
    """ Create a minimum width constraint for all polygons.
    :param polys:
    :param min_width:
    :return: Returns constraint formula.
    """
    constraints = (min_width_poly(p, min_width) for p in polys)
    return And(*constraints)


def min_notch(edges: List[OEdge],
              min_distance: int,
              min_distance_other_direction: int = None) -> FNode:
    """ Get a pySMT formula which constrains the shape to a minimum notch width.
    """

    return min_width(edges, min_distance=min_distance, min_distance_other_direction=min_distance_other_direction,
                     notch_instead_width=True)


def min_notch_poly(poly: SOPolygon, min_notch: int) -> FNode:
    """
    Get a minimum notch constraint for a symbolic polygon.
    :param polys:
    :param min_notch:
    :return:
    """
    return And(
        min_width(poly.horizontals(), min_distance=min_notch, notch_instead_width=True),
        min_width(poly.verticals(), min_distance=min_notch, notch_instead_width=True)
    )


def min_notch_of_polygons(polys: Iterable[SOPolygon], min_notch: int) -> FNode:
    """ Create a minimum notch constraint for all polygons.
    :param polys:
    :param min_notch:
    :return: Returns constraint formula.
    """
    constraints = (min_notch_poly(p, min_notch) for p in polys)
    return And(*constraints)


def preserve_absolute_edge_order_of_polygons(polys: Iterable[SOPolygon], min_distance: int = 0) -> FNode:
    """ Create a constraint that asserts the ordering of all vertical edges in x direction and all horizontal
    edges in y direction.
    :param polys:
    :param min_distance:
    :return:
    """
    # Get all edges
    polys = list(polys)
    horizontals = list(chain(*[p.horizontals() for p in polys]))
    verticals = list(chain(*[p.verticals() for p in polys]))

    return And(
        preserve_absolute_edge_order(horizontals, min_distance=min_distance),
        preserve_absolute_edge_order(verticals, min_distance=min_distance)
    )


def edge_projections_overlap(e1: OEdge, e2: OEdge, margin: int = 0) -> FNode:
    """
    Check if two edges of the same orientation over lap when projected onto the axis with the same orientation.
    :param e1:
    :param e2:
    :return:
    """

    a1, a2 = e1.interval
    b1, b2 = e2.interval
    if margin > 0:
        a1 = a1 - margin
        a2 = a2 + margin

    no_overlap = Or(
        (a1 < b1) & (a1 < b2) & (a2 < b1) & (a2 < b2),
        (b1 < a1) & (b1 < a2) & (b2 < a1) & (b2 < a2),
    )
    has_overlap = Not(no_overlap)
    return has_overlap


def preserve_relative_edge_distance(edges: List[OEdge]) -> FNode:
    """ Get a pySMT formula which constrains the edges to preserve their order.
    """

    if len(edges) > 0:
        for e in edges:
            assert e.orientation.is_vertical() == edges[0].orientation.is_vertical(), Exception(
                "Cannot mix edge orientations.")

    # Sort by original offset
    s = sorted(edges, key=lambda e: e.attrs['orig_offset'])

    constraints = [
        (b.offset - a.offset).Equals(b.attrs['orig_offset'] - a.attrs['orig_offset'])
        for a, b in window(s, 2)
    ]

    return And(*constraints)


def preserve_shape(poly: SOPolygon) -> FNode:
    """ Get a constraint that fixes the polygon shape to its original shape.
    """
    return And(
        preserve_relative_edge_distance(poly.verticals()),
        preserve_relative_edge_distance(poly.horizontals())
    )


'''     
def minimal_unconditional_edge_distance(edges, min_distance):
        """ Get a pySMT formula which constrains the edges to have a given distance.
        """
        if len(edges) > 0:
                for e in edges:
                        assert e.orientation == edges[0].orientation, Exception("Cannot mix edge orientations.")

        constraints = [ Pow((a.offset - b.offset),2) >= min_distance**2 for a,b in permutations(edges, 2)]
                
        return And(*constraints)        
'''


def test_drc_cleaner():
    points = [(1, 0), (1, 1), (0, 1), (0, 0)]
    op = OPolygon(points)
    print(op)

    points2 = [(11, 0), (11, 1), (10, 1), (10, 0)]
    op2 = OPolygon(points)

    points2 = op.points()
    print(points2)
    print(op.verticals())

    op.verticals()[0].attrs['fixed'] = True
    sop = SOPolygon('p1', op)
    print(sop)

    sop2 = SOPolygon('p2', op2)

    s = Solver()
    s.add_assertion(sop.assert_preserved_edge_orientation())
    s.add_assertion(preserve_absolute_edge_order(sop.horizontals(), min_distance=1))
    s.add_assertion(preserve_absolute_edge_order(sop.verticals(), min_distance=1))
    s.add_assertion(preserve_relative_edge_distance(sop.verticals()))

    s.add_assertion(sop2.assert_preserved_edge_orientation())
    s.add_assertion(preserve_absolute_edge_order(sop2.horizontals(), min_distance=1))
    s.add_assertion(preserve_absolute_edge_order(sop2.verticals(), min_distance=1))

    area = Symbol('area', INT)
    s.add_assertion(Equals(area, sop.area()))

    print(s.assertions)
    sat = s.check_sat()

    print('sat = ', sat)
    if sat:
        model = s.get_model()

        solved_op = sop.to_opolygon(model)
        print(model)
        print(solved_op)
        print(solved_op.points())
        print()
    else:
        pass
        # ucore = s.get_unsat_core()
        # print(ucore)



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from .signal_router import SignalRouter
import networkx as nx
from typing import Any, Dict, Set
from itertools import product


class MultiViaRouter(SignalRouter):

    def __init__(self, router: SignalRouter, node_conflict: Dict[Any, Set]):
        """
        Wrap a detail router function such that it can create double vias.

        Parameters
        ----------

        :param router: Underlying detail router
        :param node_conflict: Dict[Node, Set[Node]]
                A mapping from a node `n` to a set of nodes that collide with `n`.
        """
        self.router = router
        self.node_conflict = node_conflict if node_conflict is not None else dict()

    def route(self, graph: nx.Graph,
              terminals,
              node_cost_fn,
              edge_cost_fn
              ) -> nx.Graph:
        return _multi_via_spanning_subtree(self.router,
                                                graph,
                                                terminals,
                                                self.node_conflict,
                                                node_cost_fn,
                                                edge_cost_fn)


def _multi_via_spanning_subtree(router: SignalRouter,
                                graph: nx.Graph,
                                terminals,
                                node_conflict: Dict, node_cost_fn,
                                edge_cost_fn):
    """ Wrap a detail router function such that it can create double vias.

    Parameters
    ----------

    router: Underlying detail router
    graph: nx.Graph
    terminals: The nodes to connect.
    node_collisions: Dict[Node, Set[Node]]
            A mapping from a node `n` to a set of nodes that collide with `n`.
    """

    S1 = router.route(graph, terminals, node_cost_fn, edge_cost_fn)

    multi_vias = set()
    for a, b in S1.edges:
        multi_via = graph[a][b].get('multi_via', 1)
        if multi_via > 1:
            multi_vias.add(((a, b), multi_via))

    # Remove edges from Graph
    if multi_vias:
        G2 = graph.copy()

        for a, b in S1.edges():
            G2.add_edge(a, b, weight=0.0001)

        for (a, b), n in multi_vias:

            a_equiv = node_conflict.get(a, [])
            b_equiv = node_conflict.get(b, [])

            for c, d in product(a_equiv, b_equiv):
                if (c, d) in G2.edges:
                    G2.remove_edge(c, d)

        # Reconnect the disconnected nodes.
        for (a, b), n in multi_vias:
            # Connect two nodes.
            path = router.route(G2, [a, b], node_cost_fn, edge_cost_fn)
            for e in path.edges():
                S1.add_edge(*e)

    return S1



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import logging
import time
from typing import Dict, List, Tuple, Union
from klayout import db
import itertools
import os

from .writer import Writer
from ..layout import layers

logger = logging.getLogger(__name__)


class Label:
    def __init__(self, layer_name: str, text_source: Tuple[int, int], polygon_source: Tuple[int, int]):
        assert isinstance(polygon_source, tuple) and len(polygon_source) == 2, "source must be a 2-tuple"
        assert isinstance(text_source, tuple) and len(polygon_source) == 2, "source must be a 2-tuple"
        self.layer_name = layer_name
        self.text_source = text_source
        self.polygon_source = polygon_source


class Layer:
    def __init__(self, layer_name: str, polygon_source: Tuple[int, int]):
        assert isinstance(polygon_source, tuple) and len(polygon_source) == 2, "source must be a 2-tuple"

        self.layer_name = layer_name
        self.polygon_source = polygon_source


def _decompose_polygon(polygon: db.Polygon, ignore_non_rectilinear: bool = False) -> List[db.Box]:
    """
    Decompose a `db.Region` of multiple `db.Polygon`s into non-overlapping rectangles (`db.Box`).
    :param polygon:
    :param ignore_non_rectilinear:
    :return:
    """
    return _decompose_region(db.Region(polygon), ignore_non_rectilinear)


def _decompose_region(region: db.Region, ignore_non_rectilinear: bool = False) -> List[db.Box]:
    """
    Decompose a `db.Region` of multiple `db.Polygon`s into non-overlapping rectangles (`db.Box`).
    :param region:
    :param ignore_non_rectilinear: If set to `True` then non-rectilinear polygons are skipped.
    :return: Returns the list of rectangles.
    """
    trapezoids = region.decompose_trapezoids_to_region()
    logger.debug("Number of trapezoids: {}".format(trapezoids.size()))
    rectangles = []
    for polygon in trapezoids.each():
        box = polygon.bbox()

        if db.Polygon(box) != polygon:
            msg = "Cannot decompose into rectangles. Something is not rectilinear!"
            if not ignore_non_rectilinear:
                logger.error(msg)
                assert False, msg
            else:
                logger.warning(msg)

        rectangles.append(box)
    return rectangles


def _format_rect(box: db.Box) -> str:
    """
    Format a rectangle as a string of the form as it is required for .mag files: "xbot ybot xtop ytop"
    :param box: The rectangle.
    :return: Returns the formatted string.
    """
    lower_left = box.p1
    upper_right = box.p2
    xbot, ybot = lower_left.x, lower_left.y
    xtop, ytop = upper_right.x, upper_right.y

    assert xbot <= xtop
    assert ybot <= ytop

    rect_str = "{} {} {} {}".format(xbot, ybot, xtop, ytop)
    return rect_str


def store_layout_to_magic_file(tech_name: str,
                               output_map: Dict[str, Union[str, List[str]]],
                               layout: db.Layout,
                               pin_geometries: Dict[str, List[Tuple[str, db.Shape]]],
                               top_cell: db.Cell,
                               output_file: str,
                               ignore_non_rectilinear: bool = False,
                               scale_factor: int = 1):
    """
    Write the cell layout to a file in the Magic (.mag) format.

    Documentation of the magic file format: http://opencircuitdesign.com/magic/manpages/mag_manpage.html
    :param layout:
    :param top_cell:
    :param output_file:
    :param ignore_non_rectilinear:
    :param gds_path: Path of the source GDS2 file. This will be written into the comment of the .mag file.
    :return:
    """

    layer_config = []
    for source_layer_name, destinations in output_map.items():
        assert isinstance(destinations, str) or isinstance(destinations, List), \
            'Destination layer must either be as string or a list of strings.'

        if not isinstance(destinations, List):
            destinations = [destinations]

        for dest_name in destinations:
            # Convert layer name into (index, datatype).
            layer_config.append(
                Layer(dest_name, layers.layermap[source_layer_name])
            )

    logger.info("Number of layers: {}".format(layout.layers()))
    logger.info("Processing cell: {}".format(top_cell.name))

    mag_lines = ["magic",
                 "# Generated by librecell",
                 "tech {}".format(tech_name),
                 "timestamp {}".format(int(time.time()))
                 ]

    mag_labels = []

    transformation = db.DCplxTrans(scale_factor)

    for layer in layer_config:
        text_layer_gds_index = layer.polygon_source
        idx = layout.layer(*text_layer_gds_index)
        layer_info = layout.get_info(idx)

        layer_number = layer_info.layer
        layer_datatype = layer_info.datatype
        assert text_layer_gds_index == (layer_number, layer_datatype)

        # Fetch geometry of layer
        region = db.Region(top_cell.shapes(idx))

        line = "<< {} >>".format(layer.layer_name)
        mag_lines.append(line)

        # Scale the region.
        region = region.transformed(transformation)

        # Convert region into rectangles.
        boxes = _decompose_region(region, ignore_non_rectilinear=ignore_non_rectilinear)
        # Format rectangle strings.
        rect_lines = ["rect {}".format(_format_rect(box)) for box in boxes]
        mag_lines.extend(rect_lines)

    port_counter = itertools.count(1)
    for pin_name, pins in pin_geometries.items():
        for layer_name, pin_shapes in pins:
            pin_region = db.Region()
            pin_region.insert(pin_shapes)
            # Convert pin shape into rectangles.
            for pin_shape in pin_region.each_merged():
                pin_shape = transformation.trans(pin_shape)
                rectangles = _decompose_polygon(pin_shape, ignore_non_rectilinear)
                rectangles_str = [_format_rect(r) for r in rectangles]

                for rect_str in rectangles_str:
                    text_orientation = 0  # 0: center
                    mag_labels.append("rlabel {} {} {} {}".format(layer_name,
                                                                  rect_str,
                                                                  text_orientation,
                                                                  pin_name))

                    # TODO: is this correct?
                    mag_labels.append("port {} se".format(next(port_counter)))

    # Appends 'labels' section if there are any.
    if len(mag_labels) > 0:
        mag_lines.append("<< labels >>")
        mag_lines.extend(mag_labels)

    mag_lines.append("<< end >>\n")

    with open(output_file, "w") as mag_file:
        logger.info("Writing MAG file: {}".format(output_file))
        mag_data = "\n".join(mag_lines)
        mag_file.write(mag_data)


class MagWriter(Writer):

    def __init__(self, tech_name: str, output_map: Dict[str, str], scale_factor: float = 1):
        """

        :param tech_name:
        :param output_map:
        :param scale_factor: Scale all coordinates by this number (rounding down to next integer).
        """
        self.tech_name = tech_name
        self.output_map = output_map
        self.scale_factor = scale_factor

    def write_layout(self,
                     layout: db.Layout,
                     pin_geometries: Dict[str, List[Tuple[str, db.Shape]]],
                     top_cell: db.Cell,
                     output_dir: str,
                     ) -> None:
        mag_file_name = '{}.mag'.format(top_cell.name)
        mag_out_path = os.path.join(output_dir, mag_file_name)

        store_layout_to_magic_file(
            self.tech_name,
            self.output_map,
            layout,
            pin_geometries,
            top_cell,
            mag_out_path,
            ignore_non_rectilinear=False,
            scale_factor=self.scale_factor)



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import networkx as nx
from typing import Any, Dict, Tuple
import logging
import importlib.machinery
import types

logger = logging.getLogger(__name__)


def spacing_graph(min_spacing: Dict[Tuple[Any, Any], int]) -> nx.Graph:
    """
    Build a spacing rule graph by mapping the minimal spacing between layer a and layer b to an edge
    a-b in the graph with weight=min_spacing.
    """
    g = nx.Graph()
    for (l1, l2), spacing in min_spacing.items():
        g.add_edge(l1, l2, min_spacing=spacing)
    return g


def load_tech_file(path, module_name='tech'):
    """
    Load a python module containing technology information.
    :param path:
    :param module_name:
    :return: Handle to the module.
    """
    logger.info('Loading tech file: %s', path)
    loader = importlib.machinery.SourceFileLoader('module_name', path)
    tech = types.ModuleType(loader.name)
    loader.exec_module(tech)
    return tech



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import networkx as nx
import logging

from itertools import count
from typing import List

from lccommon import net_util

logger = logging.getLogger(__name__)


def partition(ciruit_graph: nx.MultiGraph) -> List[nx.MultiGraph]:
    """ Find sub-graphs that are connected when ignoring supply nets.
    :param ciruit_graph: nx.MultiGraph
    :return:
    """

    logger.debug('Partitioning into connected sub-graphs.')

    # Split supply nodes.
    g = nx.MultiGraph()
    cnt = count()

    node_reverse_mapping = dict()

    def replace_supply_node(n):
        if net_util.is_power_net(n):
            new = (n, next(cnt))
        else:
            new = (n, 0)
        node_reverse_mapping[new] = n
        return new

    for a, b, key, data in ciruit_graph.edges(keys=True, data=True):
        a = replace_supply_node(a)
        b = replace_supply_node(b)
        g.add_edge(a, b, key, **data)

    logger.debug('Connected components: %d', nx.number_connected_components(g))

    connected = (g.subgraph(c) for c in nx.connected_components(g))

    # Map node labels back to original labels.
    connected = [nx.relabel_nodes(c, node_reverse_mapping, copy=True) for c in connected]

    return connected



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from ..place.place import Transistor, ChannelType

from .layers import *
from typing import Any, Dict, List, Optional, Set, Tuple
import sys

# klayout.db should not be imported if script is run from KLayout GUI.
if 'pya' not in sys.modules:
    import klayout.db as pya


class TransistorLayout:
    """ Implementations of this class are responsible for drawing a transistor to the layout.
    The function `draw()` must be implemented.
    """

    def __init__(self, abstract_transistor: Transistor, location: Tuple[int, int], distance_to_outline: int, tech):
        """
        Create the layout representation of a transistor based on the abstract transistor (netlist) a location within the cell
        and design rules.
        :param abstract_transistor: Netlist representation of the transistor.
        :param location: Location in the cell matrix.
        :param distance_to_outline: TODO: This should be put into the `tech`.
        :param tech: Technology specific designrules.
        """
        raise NotImplemented()

    def terminal_nodes(self) -> Dict[str, List[Tuple[str, Tuple[int, int]]]]:
        """
        Get point-like terminal nodes in the form `{net name: {(layer name, (x, y)), ...}}`.

        This function allows to define point-like terminals at precise locations additionally to the terminals
        defined by polygons in the layout.

        This could be used for instance if a net region does not touch any grid points. Hence it is possible to insert
        off-grid routing terminals.
        """
        return dict()

    def draw(self, shapes: Dict[Any, pya.Shapes]) -> None:
        """ Draw the TransistorLayout.

        Routing terminals must be labelled with the `'net'` property.

        Example
        =======
        To insert the gate of a transistor:

        `shapes[l_poly].insert(gate_shape).set_property('net', gate_net)`

        :param shapes: Dict[layer name, pya.Shapes]
          A dict mapping layer names to pya.Shapes.
        """
        raise NotImplemented()


class DefaultTransistorLayout(TransistorLayout):
    """ Layout representation of transistor.
    
    Contains the shapes of the transistor plus shapes that mark possible locations of contacts.
    """

    def __init__(self, abstract_transistor: Transistor, location: Tuple[int, int], tech):
        """
        Create the layout representation of a transistor based on the abstract transistor (netlist) a location within the cell
        and design rules.
        :param abstract_transistor: Netlist representation of the transistor.
        :param location: Location in the cell matrix.
        :param distance_to_outline: TODO: This should be put into the `tech`.
        :param tech: Technology specific designrules.
        """

        self.abstract_transistor = abstract_transistor
        self.location = location
        self.distance_to_outline = tech.transistor_offset_y # TODO: Simplify this.

        # Get either the ndiffusion or pdiffusion layer.
        if abstract_transistor.channel_type == ChannelType.NMOS:
            l_diffusion = l_ndiffusion
        else:
            l_diffusion = l_pdiffusion

        # Diffusion layer.
        self.l_diffusion = l_diffusion

        # Calculate minimal distance from active region to upper and lower cell boundaries.
        poly_half_spacing = (tech.min_spacing[(l_poly, l_poly)] + 1) // 2
        active_half_spacing = (tech.min_spacing[(l_diffusion, l_diffusion)] + 1) // 2
        # Distance from active to active in neighbouring cell must be kept,
        # as well as distance from poly to poly in neighbouring cell.
        min_distance_to_outline = max(active_half_spacing, tech.gate_extension + poly_half_spacing)

        assert self.distance_to_outline >= min_distance_to_outline, 'Chosen distance will violate minimum spacing rules. {} >= {}.'.format(
            self.distance_to_outline, min_distance_to_outline)

        # Bottom left of l_diffusion.
        x, y = location

        # Choose l_diffusion width such that at least one contact can be placed on each side of the transistor.
        w = tech.unit_cell_width + tech.via_size[l_diff_contact] + 2 * tech.minimum_enclosure[
            (l_diffusion, l_diff_contact)]

        h = abstract_transistor.channel_width

        x_eff = (x + 1) * tech.unit_cell_width - w // 2
        y_eff = 0
        if y % 2 == 1:
            # Top aligned.
            y_eff = y * tech.unit_cell_height - self.distance_to_outline
            # y_eff = grid_floor(y_eff, tech.routing_grid_pitch_y, tech.grid_offset_y) + tech.via_size[l_diff_contact] // 2 + \
            #         tech.minimum_enclosure[(l_diffusion, l_diff_contact)]
            y_eff = y_eff - h
        else:
            # Bottom aligned
            y_eff = y * tech.unit_cell_height + self.distance_to_outline
            # y_eff = grid_ceil(y_eff, tech.routing_grid_pitch_y, tech.grid_offset_y) - tech.via_size[l_diff_contact] // 2 - \
            #         tech.minimum_enclosure[(l_diffusion, l_diff_contact)]

        # Create shape for active layer.
        active_box = pya.Box(
            x_eff,
            y_eff,
            x_eff + w,
            y_eff + h
        )

        # Enclose active regions of PMOS transistors with l_nwell.

        # Get layer depending on channel type.
        l_well = l_nwell if abstract_transistor.channel_type == ChannelType.PMOS else l_pwell
        # Well layer.
        self.l_well = l_well

        # Get minimum overlap from tech file.
        well2active_overlap = tech.minimum_enclosure.get((l_well, l_diffusion), 0)
        if not isinstance(well2active_overlap, tuple):
            well2active_overlap = (well2active_overlap, well2active_overlap)
        well2active_overlap_x, well2active_overlap_y = well2active_overlap

        # Create shape of nwell or pwell.
        well_box = pya.Box(
            x_eff - well2active_overlap_x,
            y_eff - well2active_overlap_y,
            x_eff + w + well2active_overlap_x,
            y_eff + h + well2active_overlap_y
        )

        center_x = active_box.center().x
        assert (center_x - tech.grid_offset_x) % tech.routing_grid_pitch_x == 0, Exception(
            "Gate not x-aligned on grid.")

        source_box = pya.Box(
            x_eff,
            y_eff,
            center_x - tech.gate_length // 2,
            y_eff + h
        )

        drain_box = pya.Box(
            center_x - tech.gate_length // 2,
            y_eff,
            x_eff + w,
            y_eff + h
        )

        top = active_box.top
        bottom = active_box.bottom

        gate_top = top + tech.gate_extension
        gate_bottom = bottom - tech.gate_extension

        # Create gate terminals.
        terminals = {
            abstract_transistor.gate_net: [
                (l_poly, (center_x, gate_top)),
                (l_poly, (center_x, gate_bottom))
            ]
        }

        # Create gate shape.
        gate_path = pya.Path.new([
            pya.Point(center_x, gate_top),
            pya.Point(center_x, gate_bottom)],
            tech.gate_length,
            0,
            0)

        self._gate_path = gate_path
        self._active_box = active_box
        self._source_box = source_box
        self._drain_box = drain_box
        self._terminals = terminals
        self._well_box = well_box

    def terminal_nodes(self) -> Dict[str, List[Tuple[str, Tuple[int, int]]]]:
        """
        Get point-like terminal nodes in the form `{net name: {(layer name, (x, y)), ...}}`.
        """
        return self._terminals

    def draw(self, shapes: Dict[Any, pya.Shapes]) -> None:
        """ Draw a TransistorLayout.

        :param shapes: Dict[layer name, pya.Shapes]
          A dict mapping layer names to pya.Shapes.
        """

        # Create well and active shape.
        shapes[self.l_well].insert(self._well_box)
        shapes[self.l_diffusion].insert(self._active_box)
        shapes[self.l_diffusion].insert(self._source_box).set_property('net', self.abstract_transistor.source_net)
        shapes[self.l_diffusion].insert(self._drain_box).set_property('net', self.abstract_transistor.drain_net)

        # Create gate shape.
        inst = shapes[l_poly].insert(self._gate_path)
        inst.set_property('net', self.abstract_transistor.gate_net)



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import networkx as nx
from .signal_router import SignalRouter

from pysmt.shortcuts import *
from pysmt.fnode import FNode
from pysmt.solvers.z3 import Z3Model, Z3Converter
from typing import Any, Dict, List
from itertools import product, chain
import logging

# Import z3 if package is available.
try:
    import z3

    _z3_available = True
except ImportError:
    _z3_available = False

logger = logging.getLogger(__name__)


class MIPSignalRouter(SignalRouter):

    def __init__(self):
        pass

    def route(self, G: nx.Graph,
              terminals: List[Any],
              node_cost_fn,
              edge_cost_fn
              ) -> nx.Graph:
        return min_steiner_tree(
            G,
            [terminals],
            node_cost_fn,
            edge_cost_fn,
        )[0]


class MIPGraphRouter():

    def route(self, G: nx.Graph,
              terminals: List[Any],
              node_cost_fn,
              edge_cost_fn
              ) -> nx.Graph:
        return min_steiner_tree(
            G,
            terminals,
            node_cost_fn,
            edge_cost_fn,
        )


def min_steiner_tree(
        G: nx.Graph,
        signals: List[List[Any]],
        node_cost_fn,
        edge_cost_fn,
) -> nx.Graph:
    """
    Find a minimum weight Steiner tree which connects all `terminals` in `G`.

    :param G: Graph.
    :param terminals: List of terminals to be connected.

    References:
    Sunil Chopra and Chih-Yang Tsai, Polyhedral approaches for the Steiner Tree on graphs, 2000.
    """

    optimize = False
    solver_name = 'z3'
    debug = False

    if optimize:
        assert _z3_available, "Need z3-solver for optimizations."

    if optimize and solver_name != 'z3':
        logger.warning("Can not use {} for optimizations. Switching to z3.".format(solver_name))
        solver_name = 'z3'

    # Get solver instance.
    if debug:
        logger.debug("Use UnsatCoreSolver.")
        solver = UnsatCoreSolver(name=solver_name)
    else:
        solver = Solver(name=solver_name)

    # Create optimizer instances if needed.
    if optimize:
        optimizer = z3.Optimize()
        converter = solver.converter
        logger.info("SMT optimizer: z3")
    else:
        optimizer = None
        converter = None
        logger.info("SMT solver: {}".format(type(solver).__name__))

    def add_assertion(assertion: FNode, **kwargs):
        """
        Add assertion to solver and optimizer.
        :param assertion:
        :param kwargs:
        :return:
        """

        assert isinstance(assertion, FNode)
        solver.add_assertion(assertion, **kwargs)
        if optimizer:
            optimizer.add(
                converter.convert(assertion)
            )

    def minimize(objective: FNode):
        """
        Add minimization objective to optimizer.
        :param objective:
        :return:
        """
        if optimizer:
            optimizer.minimize(
                converter.convert(objective)
            )

    def create_constraints(source, sinks) -> Dict[Any, FNode]:
        """
        Create constraints for steiner tree of a single signal.
        :param source:
        :param sinks:
        :return:
        """
        # Edge choice variables.
        x = {(i, j): FreshSymbol(INT, template='x%d')
             for i, j in G.edges()
             }

        # Flow variables
        f = dict()
        f.update({(k, i, j): FreshSymbol(INT, template='f%d')
                  for k, (i, j) in product(sinks, G.edges)
                  })
        f.update({(k, j, i): FreshSymbol(INT, template='f%d')
                  for k, (i, j) in product(sinks, G.edges)
                  })

        y = dict()
        y.update({(i, j): FreshSymbol(INT, template='y%d') for i, j in G.edges})
        y.update({(j, i): FreshSymbol(INT, template='y%d') for i, j in G.edges})

        # Constraint (1)
        for k in sinks:
            for j in G.nodes:
                left = Int(0) + sum((f[(k, i, j)] - f[(k, j, i)] for i in G.neighbors(j)))

                if j == source:
                    right = -1
                elif j == k:
                    right = 1
                else:
                    right = 0

                add_assertion(left.Equals(Int(right)))

        # Constraint (3)
        for k, (i, j) in product(sinks, G.edges):
            add_assertion(f[(k, i, j)] <= y[(i, j)])
            add_assertion(f[(k, j, i)] <= y[(j, i)])

        # Constraint (4)
        for i, j in G.edges:
            y_ij = y[(i, j)]
            y_ji = y[(j, i)]
            x_ij = x[(i, j)]
            add_assertion(
                Equals(y_ij + y_ji, x_ij)
            )

        # f_kij must be >= 0
        for f_kij in f.values():
            add_assertion(f_kij >= 0)

        return x

    def create_node_usage_indicator(edge_usage: Dict[Any, FNode]) -> Dict[Any, FNode]:
        """
        Create variables indicating if a node is used based on usage of edges.
        :param edge_usage:
        :return:
        """
        # Introduce variable for indicating node usage.
        node_used = {n: FreshSymbol(INT, template='node_used_%d')
                     for n in G.nodes}
        # node_used can be either 0 or 1
        for n, var in node_used.items():
            add_assertion(var >= 0)
            add_assertion(var <= 1)
        # node_used must be 1 if an adjacent edge is used.
        for i, j in G.edges:
            add_assertion(
                node_used[i] >= edge_usage[(i, j)]
            )
            add_assertion(
                node_used[j] >= edge_usage[(i, j)]
            )

        return node_used

    total_cost = Int(0)

    xs = []
    if len(signals) > 1:
        node_sharing = {n: 0 for n in G.nodes}
    else:
        node_sharing = dict()

    for terminals in signals:
        source = terminals[0]
        sinks = terminals[1:]
        x = create_constraints(source, sinks)
        xs.append(x)

        node_used = create_node_usage_indicator(x)

        # Count by how many signals a node is used.
        if len(signals) > 1:
            for n, var in node_used.items():
                node_sharing[n] += var

        if node_cost_fn is not None:
            # Node weights.

            node_weight = sum(
                Int(node_cost_fn(n)) * node_used[n]
                for n in G.nodes
            )

            total_cost += node_weight

        if edge_cost_fn is not None:
            edge_weight = sum(
                # Int(data.get('weight', 0)) * x[(i, j)]
                Int(edge_cost_fn((i, j))) * x[(i, j)]
                for i, j, data in G.edges(data=True)
            )

            total_cost += edge_weight

    # Distinct nodes constraint.
    # Assert that a node cannot be shared among signals.
    for n, num_signals in node_sharing.items():
        add_assertion(num_signals <= 1)

    # Minimize sum of edge weights.
    minimize(total_cost)

    if optimizer:
        logger.info('Invoke SMT optimizer')
        sat = optimizer.check()
        logger.info('SMT optimize result: %s', 'SAT' if sat else 'UNSAT')
    else:
        logger.info('Invoke SMT solver')
        sat = solver.check_sat()
        logger.info('SMT result: %s', 'SAT' if sat else 'UNSAT')

    if sat:
        if optimizer:
            model = Z3Model(solver.environment, optimizer.model())
        else:
            model = solver.get_model()

        # Map solution back to a routing graph.
        routing_trees = []
        for x in xs:
            routing_tree = nx.Graph()
            for i, j in G.edges:
                x_ij = x[(i, j)]
                value = model.get_py_value(x_ij)
                if value == 1:
                    routing_tree.add_edge(i, j)
            routing_trees.append(routing_tree)

        return routing_trees

    else:
        msg = 'UNSAT: Constraints not satisfiable.'
        logger.error(msg)
        if debug:
            core = solver.get_named_unsat_core()
            logger.info('unsat core: %s', list(core.keys()))
            logger.debug('unsat core: %s', core)

        raise Exception(msg)


def test_mip_route():
    logging.basicConfig(level=logging.INFO)
    import matplotlib.pyplot as plt

    G = nx.Graph()

    num_x = 10
    num_y = 10
    x = range(0, num_x)
    y = range(0, num_y)

    # Store drawing positions of vertices. For plotting only.
    pos = {}

    # Construct mesh
    for i, (x, y) in enumerate(product(x, y)):
        G.add_node((x, y))
        pos[(x, y)] = (x, y)

        if x < num_x - 1 and not (1 <= y < 5 and x == 4):
            G.add_edge((x, y), (x + 1, y), weight=1, orientation='h')

        if y < num_y - 1:
            G.add_edge((x, y), (x, y + 1), weight=1, orientation='v')

    # Draw mesh
    nx.draw_networkx(G, pos=pos, node_color='gray', node_size=8, edge_color='lightgray', hold=True)

    signals = [  # [terminals, ...]
        [(0, 0), (8, 5), (7, 7), (6, 3)],
        [(1, 1), (9, 0)],
        #[(3, 3), (3, 6)],
        #[(0, 9), (9, 0)],
        # [(0,1), (9,2)],
        # [(1,1), (8,9), (7,4)],
        # [(1,1), (2,0)],
        # [(4,1), (5,1), (7,3)],
        # [(1,3), (4,8)],
        # [(3,3), (6,5)],
        # [(1,2), (0,0), (0,4), (2,0), (2,4), (9,9), (5,5)],
        # [(1,1), (1,8)],
        # [(3,1), (3,8)],
        # [(9,0), (0,9), (5,5)],
    ]

    routing_trees = min_steiner_tree(G, signals,
                                     node_cost_fn=None,
                                     edge_cost_fn=lambda e: 1)

    colors = ['red', 'blue', 'green', 'orange', 'violet']
    # routing_trees = route_hv(DijkstraRouter(), G, signals, orientation_change_penalty=10)

    # edge_labels = {(a, b): "%.2f" % data.get('weight', 0) for (a, b, data) in G.edges(data=True)}

    # nx.draw_networkx_edges(G, pos, edgelist=routing_tree.edges, width=4, edge_color=colors[0])

    for i, signal in enumerate(signals):
        edges = list(routing_trees[i].edges)
        nx.draw_networkx_edges(G, pos, edgelist=edges, width=4, edge_color=colors[i])

    nx.draw_networkx_nodes(G, pos, nodelist=list(chain(*signals)), node_size=32, node_color='black')
    # nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels)

    plt.draw()
    plt.show()



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from pysmt.shortcuts import *
import re

from pysmt.environment import get_env

"""
Collection of theorem proofs.
Just a scratch pad. This code is not directly used in lclayout. 
"""


def proof(f):
    return is_valid(f)


def is_valid(f):
    """ Proof the validity of a formula.
    """
    s = Solver()
    s.add_assertion(Not(f))

    has_counter_example = s.check_sat()

    if has_counter_example:
        m = s.get_model()
        print(m)

    return not has_counter_example


def is_equal(f1, f2):
    return proof(
        Iff(f1, f2)
    )


def symbols(names, dtype):
    return [Symbol(n, dtype) for n in re.split(',\s*|\s+', names)]


def _init_env():
    """
    Initialize pySMT environment at beginning of each test.
    """
    pysmt.shortcuts.reset_env()
    get_env().enable_infix_notation = True


def test_proof_de_morgan():
    """ Proof de Morgans law.
    """
    _init_env()
    b1, b2 = symbols('b1, b2', BOOL)

    f = b1 & b2
    demorgan = Not(b1) | Not(b2)

    theorem = Iff(Not(f), demorgan)

    assert proof(theorem)


def test_proof_interval_overlap_formula():
    """ Proof interval overlap formula.
    Proof that f2 checks if two intervals overlap given a1<=a2, b1<=b2.
         a1                 a2
          |-----------------|
      b1          b2
       |----------|
    """
    _init_env()

    a1, a2, b1, b2 = symbols('a1, a2, b1, b2', INT)

    # To proove: This should be sufficient to detect if intervals overlap given a1<=a2, b1<=b2.
    f2 = Or(
        (a1 <= b1) & (b1 <= a2),
        (b1 <= a1) & (a1 <= b2),
    )

    """
    f3 = Or(
        (a1 <= b1) & (b1 <= a2),
        (a1 <= b2) & (b2 <= a2),
        (b1 <= a1) & (a1 <= b2)
    )
    
    f4 = Or(
        (a1 <= b1) & (b1 <= a2),
        (a1 <= b2) & (b2 <= a2),
        (b1 <= a1) & (a1 <= b2),
        (b1 <= a2) & (a2 <= b2),
    )
    """

    # Reference.
    # Check for strict ordering of a and b.
    # The intervals don't overlap iff a* < b* OR b* < a*
    no_overlap = Or(
        (a1 < b1) & (a1 < b2) & (a2 < b1) & (a2 < b2),
        (b1 < a1) & (b1 < a2) & (b2 < a1) & (b2 < a2),
    )
    has_overlap = Not(no_overlap)

    # Assume edges are oriented in positive direction.
    assume_edge_direction = (a1 <= a2) & (b1 <= b2)

    # assume_edge_direction => (f2 <-> has_overlap)
    theorem = assume_edge_direction.Implies(Iff(f2, has_overlap))

    assert proof(theorem)



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import klayout.db as pya
from klayout.db import Point, Shape, Polygon, Shapes, Region

from . import base as cleaner
from .base import OPolygon, SOPolygon, Orientation, OEdge

from itertools import count, product, combinations
from typing import Dict, List, Iterable, Set, Tuple, Union

import logging

from .. import tech_util

from pysmt.shortcuts import *
from pysmt.fnode import FNode
from pysmt.solvers.z3 import Z3Model, Z3Converter

import math
from ..layout.layers import *

logger = logging.getLogger(__name__)


def _polygon_to_points(poly) -> List[Tuple[int, int]]:
    """
    Convert a klayout polygon to a list of (x,y) tuples.
    The ordering of the points is reversed.
    :param poly:
    :return:
    """
    if poly is None:
        return []
    else:
        return [(p.x, p.y) for p in reversed(list(poly.each_point()))]


def _shape_to_points(shape: pya.Shape) -> List[Tuple[int, int]]:
    """ Convert a pya.Shape to a list of (x,y) tuples.
    """
    poly = shape.simple_polygon
    return _polygon_to_points(poly)


def _points_to_shape(points: Iterable[Tuple[int, int]]) -> pya.Polygon:
    """
    Create a klayout Polygon from a list of (x,y) tuples.
    :param points:
    :return:
    """
    return pya.Polygon.new([pya.Point(x, y) for x, y in points])


def clean(tech,
          shapes: Dict[str, pya.Shapes],
          white_list: Set[Tuple[str, pya.Shape]] = None,
          enable_min_area=False,
          optimize: bool = False,
          debug=False,
          solver_name: str = 'z3'):
    """ Performs DRC cleaning on the `shapes`.

    :param tech: Tech module. Defines the design rules.
    :param shapes:
    :param white_list: Set[Tuple[layer name, shape])]
        If this is used, then only the shapes in the white list will be modified.
    :param enable_min_area: Enable minimum area constraint. Disabled by default.
        Minimum area constraint showed to slow things down because they are not modelled as a integer linear program (not linear).
    :param optimize: Use z3 to find an optimal solution.
        Shape sizes are minimized.
        This will overwrite the solver name because optimizations are currently only implemented for z3.
    :param debug: Use an `UnsatCoreSolver` for debugging of unsatisfiable constraints.
        No optimizations will be performed in this mode.
    :param solver_name: Name of the SMT solver to be used.
    :return: Returns `True` iff DRC cleaning was successful.
    """

    if debug and optimize:
        logger.warning("Optimizations disabled in debug unsat core mode.")
        optimize = False

    if optimize:
        import z3

    if optimize and solver_name != 'z3':
        logger.warning("Can not use {} for optimizations. Switching to z3.".format(solver_name))
        solver_name = 'z3'

    # Graph representing the spacing between layers. Node: layer, Edge['min_spacing']: spacing rule between layers.
    spacing_graph = tech_util.spacing_graph(tech.min_spacing)

    _sympoly_map = dict()
    _counter = count()

    def get_sympoly(shape: Union[pya.Shape, pya.Polygon], layer: str) -> SOPolygon:
        """ Given a pya.Shape get the symbolic polygon.
        """

        assert isinstance(shape, pya.Shape) or isinstance(shape, pya.Polygon)

        if isinstance(shape, pya.SimplePolygon):
            poly = shape
        elif isinstance(shape, pya.Polygon):
            poly = shape.to_simple_polygon()
        else:
            poly = shape.simple_polygon

        key = (layer, poly)

        if key not in _sympoly_map:
            points = _shape_to_points(shape)
            opoly = OPolygon(points=points)
            sympoly = SOPolygon('%s_%s' % (layer, next(_counter)), opoly)
            _sympoly_map[key] = sympoly

        return _sympoly_map[key]

    # Convert all polygons into symbolic polygons.
    logger.debug("Convert polygons into symbolic polygons.")

    sym_polys = {
        layer: [get_sympoly(s, layer) for s in ss.each()]
        for layer, ss in shapes.items()
    }

    # Get set of polygons that are allowed to be modified.
    if white_list is not None:
        white_list = {
            (layer, get_sympoly(shape=shape, layer=layer))
            for layer, shape in white_list
        }
    else:
        white_list = set()

    if len(white_list) == 0:
        logger.warning("No white-listed polygons. Will not perform any cleaning.")

    # Get solver instance.
    if debug:
        logger.debug("Use UnsatCoreSolver.")
        solver = UnsatCoreSolver(name=solver_name)
    else:
        solver = Solver(name=solver_name)

    # Create optimizer instances if needed.
    if optimize:
        optimizer = z3.Optimize()
        converter = solver.converter
        logger.info("SMT optimizer: z3")
    else:
        optimizer = None
        converter = None
        logger.info("SMT solver: {}".format(type(solver).__name__))

    def add_assertion(assertion: FNode, **kwargs):
        """
        Add assertion to solver and optimizer.
        :param assertion:
        :param kwargs:
        :return:
        """

        assert isinstance(assertion, FNode)
        solver.add_assertion(assertion, **kwargs)
        if optimizer:
            optimizer.add(
                converter.convert(assertion)
            )

    def minimize(objective: FNode):
        """
        Add minimization objective to optimizer.
        :param objective:
        :return:
        """
        if optimizer:
            optimizer.minimize(
                converter.convert(objective)
            )

    def maximize(objective: FNode):
        """
        Add maximization objective to optimizer.
        :param objective:
        :return:
        """
        if optimizer:
            optimizer.maximize(
                converter.convert(objective)
            )

    # WIP: detecting gate edges to preserve gates
    # def get_interacting_edges(a: SOPolygon, b: SOPolygon) -> List[OEdge]:
    #     """
    #     Find edges of `a` that interact with polygon `b`.
    #     Operates on original polygon.
    #     :param a:
    #     :param b:
    #     :return: List of OEdges that interact with b.
    #     """
    #
    #     reverse_map = dict()
    #     for orig_edge, sym_edge in zip(a.original_opolygon.edges, a.edges):
    #         (x1, y1), (x2, y2) = orig_edge.endpoints
    #         reverse_map[((x1, y1), (x2, y2))] = sym_edge
    #         reverse_map[((x2, y2), (x1, y1))] = sym_edge
    #
    #     poly_a = pya.Polygon(a.original_opolygon.points())
    #     poly_b = pya.Polygon(b.original_opolygon.points())
    #
    #     # Get edges of a that interact with b.
    #     interacting = pya.Edges(poly_a).interacting(pya.Region(poly_b))
    #
    #     # Convert back to corresponding symbolic edges.
    #     interacting_symbolic_edges = [reverse_map[((e.x1, e.y1), (e.x2, e.y2))]
    #                                   for e in interacting
    #                                   ]
    #     return interacting_symbolic_edges
    #
    # for active_shape in shapes[l_active].each():
    #     # Find touching poly-silicon shapes
    #     active_region = pya.Shapes()
    #     active_region.insert(active_shape)
    #     interacting = pya.Region(shapes[l_poly]).interacting(pya.Region(active_region))
    #     for i in interacting:
    #         gate_edges = get_interacting_edges(get_sympoly(i, l_poly),
    #                                            get_sympoly(active_shape, l_active))

    # Add assertions

    # Fix all polygons that are not in the white list.
    fix_counter = 0
    for layer, polys in sym_polys.items():
        for poly in polys:
            if (layer, poly) not in white_list:
                add_assertion(poly.assert_absolute_fixed(),
                              named='fixed shapes ({})'.format(layer))
                fix_counter += 1

    logger.debug("Fixed polygons: {}".format(fix_counter))
    logger.debug("Free polygons: {}".format(len(white_list)))

    # Fix via shapes but allow them to be moved.
    logger.debug("Assuming immutable via shapes.")
    via_layers = set(tech.via_layers.nodes())
    logger.debug("Add constraint for relative immutable shapes: {}".format(via_layers))
    relative_fixed_layers = via_layers
    for l in relative_fixed_layers:
        for poly in sym_polys[l]:
            add_assertion(cleaner.preserve_shape(poly),
                          named='relative immutable shapes ({})'.format(l))

    # Allow free polygons to be changed in shape but the underlying structure must be preserved.
    fixed_form_layers = shapes.keys()
    for l in fixed_form_layers:
        for poly in sym_polys[l]:
            add_assertion(poly.assert_preserved_edge_orientation(),
                          named='fixed edge orientation ({})'.format(l))

    # TODO: more generic, based on technology file
    containtement_constraints = [
        # Syntax ([Container shapes], [shapes that must be inside the container shape, ...])
        ([l_abutment_box], [l_ndiffusion, l_pdiffusion, l_diff_contact, l_poly_contact, l_via1, l_poly, l_metal1, l_metal2]),
        ([l_ndiffusion], [l_diff_contact]),
        ([l_pdiffusion], [l_diff_contact]),
        ([l_poly], [l_poly_contact]),
        ([l_metal1], [l_poly_contact, l_diff_contact, l_via1]),
        ([l_metal2], [l_via1]),
    ]

    def assert_inside_naive(outer_poly: SOPolygon, inner_poly: SOPolygon, min_inside=1) -> FNode:
        """ Create a constraint that asserts that `inner_poly` is contained in `outer_poly`.
        Requirement: The original polygons are required to fulfill this property already!

        :param outer_poly: Container polygon.
        :param inner_poly: Inner polygon.
        :param min_inside: Minimum distance of the inner polygon to the container.
        :return: FNode
        """

        v1 = outer_poly.verticals()
        v2 = inner_poly.verticals()

        h1 = outer_poly.horizontals()
        h2 = inner_poly.horizontals()

        constraints = []

        for v in v1:
            for u in v2:
                if v.attrs['orig_offset'] < u.attrs['orig_offset']:
                    constraints.append(u.offset - v.offset >= min_inside)
                elif v.attrs['orig_offset'] > u.attrs['orig_offset']:
                    constraints.append(v.offset - u.offset >= min_inside)

        for h in h1:
            for i in h2:
                if h.attrs['orig_offset'] < i.attrs['orig_offset']:
                    constraints.append(i.offset - h.offset >= min_inside)
                elif h.attrs['orig_offset'] > i.attrs['orig_offset']:
                    constraints.append(h.offset - i.offset >= min_inside)

        return And(*constraints)

    min_enclosure = dict()

    # Add half-spacing rules.
    # Half-spacing rules assert that spacing rules are hold when abutting two cells.
    # The approach is to enforce a spacing to the cell boundary.
    logger.debug("Add half-spacing constraints.")
    for l1, l2, data in spacing_graph.edges(data=True):

        half_spacing = (data['min_spacing'] + 1) // 2

        outer = l_abutment_box
        for inner in [l1, l2]:
            min_enc = min_enclosure.get((outer, inner), 0)
            min_enclosure[(outer, inner)] = max(min_enc, half_spacing)

    # Add minimum enclosure rules (of vias).
    for [l1, l2], via_dict in tech.via_layers.edges.items():
        via_layer=via_dict['layer']
        for outer in (l1, l2):
            min_enclosure[(outer, via_layer)] = tech.minimum_enclosure[(outer, via_layer)]

    for outer_layers, inner_layers in containtement_constraints:
        for outer_layer, inner_layer in product(outer_layers, inner_layers):

            min_inside = min_enclosure.get((outer_layer, inner_layer), 1)

            for outer_shape in shapes[outer_layer].each():
                # logger.debug("outer_shape = %s", outer_shape)
                _outer = Shapes()
                _outer.insert(outer_shape)
                _outer = Region(_outer)
                inner = Region(shapes[inner_layer])

                inside = inner.inside(_outer)
                outer_poly = get_sympoly(outer_shape, outer_layer)
                for inner_shape in inside.each():
                    inner_poly = get_sympoly(inner_shape, inner_layer)

                    if (outer_layer, outer_poly) in white_list or (inner_layer, inner_poly) in white_list:
                        # Add constraint only if both polygons are modifiable.
                        add_assertion(
                            assert_inside_naive(outer_poly, inner_poly, min_inside=min_inside),
                            named="min inside {}-{}".format(outer_layer, inner_layer)
                        )

    # Min width
    logger.debug("Add minimum width constraints.")
    for layer, min_width in tech.minimum_width.items():
        for p in sym_polys[layer]:
            add_assertion(
                cleaner.min_width_of_polygons([p], min_width=min_width),
                named="min width ({})".format(layer)
            )

    # Min notch
    logger.debug("Add minimum notch constraints.")
    for layer, min_notch in tech.minimum_notch.items():
        for p in sym_polys[layer]:
            add_assertion(
                cleaner.min_notch_of_polygons([p], min_notch=min_notch),
                named="min notch ({})".format(layer)
            )

    # Min area
    if enable_min_area:
        logger.debug("Add minimum area constraints.")
        # logger.warning('Enabled min_area constraints. (Results in non-linear problem and can slow things down.)')
        for layer, min_area in tech.min_area.items():
            for p in sym_polys[layer]:
                if len(p.points()) == 4 and (layer, p) in white_list:
                    if min_area > 0:
                        # add_assertion(
                        #     p.area() >= min_area,
                        #     named="min area {}".format(layer)
                        # )

                        # Create a square shape.
                        side = int(math.sqrt(min_area) + 1)
                        for e in p.edges:
                            add_assertion(
                                e.length() >= side if optimize else Equals(e.length(), Int(side)),
                                named="min area {}".format(layer)
                            )

    # Min spacing
    logger.debug('Add minimum spacing rules.')
    # TODO:
    logger.warning("Fixed shapes are excluded from spacing rules.")
    for (layer1, layer2), spacing in tech.min_spacing.items():

        min_dist_v = spacing
        min_dist_h = min_dist_v

        if layer1 == layer2:
            # Intra-layer spacing
            pairs = combinations(sym_polys[layer1], 2)
        else:
            # Inter-layer spacing
            pairs = product(sym_polys[layer1], sym_polys[layer2])

        for p1, p2 in pairs:

            # Don't add constraints if both polygons are immutable.
            if (layer1, p1) not in white_list and (layer2, p2) not in white_list:
                # TODO: include fixed layers? (implicit DRC check)
                continue

            for e1, e2 in product(p1.verticals(), p2.verticals()):
                add_assertion(
                    Implies(cleaner.edge_projections_overlap(e1, e2, margin=min_dist_h),
                            cleaner.preserve_absolute_edge_order([e1, e2], min_distance=min_dist_v)),
                    named='min spacing horizontal {}-{}'.format(layer1, layer2)
                )

            for e1, e2 in product(p1.horizontals(), p2.horizontals()):
                add_assertion(
                    Implies(cleaner.edge_projections_overlap(e1, e2, margin=min_dist_v),
                            cleaner.preserve_absolute_edge_order([e1, e2], min_distance=min_dist_h)),
                    named='min spacing vertical {}-{}'.format(layer1, layer2)
                )

    if optimizer:
        # Optimization objectives
        for layer, polys in sym_polys.items():
            for poly in polys:
                for edge in poly.edges:
                    # TODO: add constraints sorted by original edge length descending.
                    # -> better area optimization because objectives are priorized lexicographically.
                    if edge.orientation in [Orientation.UP, Orientation.LEFT]:
                        # Upper or right edge.
                        minimize(edge.offset)
                    else:
                        # Lower or left edge.
                        maximize(edge.offset)

        logger.info('Invoke SMT optimizer')
        sat = optimizer.check()
        logger.info('SMT optimize result: %s', 'SAT' if sat else 'UNSAT')
    else:
        logger.info('Invoke SMT solver')
        sat = solver.check_sat()
        logger.info('SMT result: %s', 'SAT' if sat else 'UNSAT')

    if sat:
        if optimizer:
            model = Z3Model(solver.environment, optimizer.model())
        else:
            model = solver.get_model()

        # Convert back to constant polygons.
        solutions = {
            l: [p.to_opolygon(model) for p in polys]
            for l, polys in sym_polys.items()
        }

        new_shapes = {
            l: [_points_to_shape(poly.points()) for poly in polys]
            for l, polys in solutions.items()
        }

        # Update shapes with solutions
        for l, new_s in new_shapes.items():
            s = shapes[l]
            s.clear()
            s.insert(pya.Region(new_s))

        return True

    else:
        logger.error('UNSAT: Constraints not satisfiable.')
        if debug:
            core = solver.get_named_unsat_core()
            logger.info('unsat core: %s', list(core.keys()))
            logger.debug('unsat core: %s', core)
        return False
